import json
from collections import namedtuple
from functools import wraps

from requests import Session

from . import exceptions
from .logging import get_logger

# Set up logger for this module
logger = get_logger("session")


def handle_response(func):
    """
    Decorator which parses the response returned
    by Moku API Server
    """

    @wraps(func)
    def func_wrapper(self, *args, **kwargs):
        response = func(self, *args, **kwargs)
        return self.resolve(response)

    return func_wrapper


class RequestSession:
    "Base HTTP Requests class"

    json_headers = {"Content-type": "application/json"}
    sk_name = "Moku-Client-Key"  # session key name

    def __init__(self, ip, connect_timeout, read_timeout, **kwargs):
        self.ip_address = ip
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout
        self.rs = Session()
        logger.debug(
            f"Session initialized for {ip} with timeouts: connect={connect_timeout}s, read={read_timeout}s"
        )

        # support arbitrary session arguments
        for k, v in kwargs.items():
            if k.lower().startswith("session_"):
                k = k.split("session_")[1]
                setattr(self.rs, k, v)

    def update_sk(self, response):
        key = response.headers.get(self.sk_name)
        if key:
            self.session_key = key
            self.rs.headers.update({self.sk_name: key})
            logger.debug(
                f"Session key updated: {key[:8]}..."
                if len(key) > 8
                else f"Session key updated: {key}"
            )

    def url_for(self, group, operation):
        return f"http://{self.ip_address}/api/{group}/{operation}"

    def url_for_v2(self, location):
        return f"http://{self.ip_address}/api/v2/{location}"

    def timeout_headers(self, rt_increase=0):
        "Returns timeout headers required for http request"
        return tuple([self.connect_timeout, self.read_timeout + rt_increase])

    @handle_response
    def get(self, group, operation):
        "Executes get call and returns the response"
        url = self.url_for(group, operation)
        logger.debug(f"GET {url}")
        response = self.rs.get(url, timeout=self.timeout_headers())
        logger.debug(f"GET {url} - Status: {response.status_code}")
        return response

    @handle_response
    def post(self, group, operation, params=None):
        "Executes post call and returns the response"
        # As get_data has an explicit read_timeout parameter,
        # it should be considered applicable cases, in all other
        # cases default it to 0
        _timeout = None
        if params:
            _timeout = params.get("timeout", 0)
        _to_inc = 0 if not _timeout else _timeout
        url = self.url_for(group, operation)
        logger.debug(f"POST {url} with params: {params}")
        response = self.rs.post(
            url,
            json=params,
            timeout=self.timeout_headers(_to_inc),
            headers=self.json_headers,
        )
        logger.debug(f"POST {url} - Status: {response.status_code}")
        return response

    @handle_response
    def post_raw_json(self, group, operation, data):
        "Executes post call and returns the response"
        return self.rs.post(
            self.url_for(group, operation),
            json=data,
            headers=self.json_headers,
        )

    def post_to_v2_raw(self, location, params=None):
        "Executes post call to api v2 and returns the response"
        response = self.rs.post(self.url_for_v2(location), json=params)
        return response

    def post_to_v2(self, location, params=None):
        url = self.url_for_v2(location)
        logger.debug(f"POST v2 {url} with params: {params}")
        response = self.rs.post(url, json=params)
        logger.debug(f"POST v2 {url} - Status: {response.status_code}")
        if response.status_code != 200:
            logger.error(f"API v2 request failed with status {response.status_code}")
            raise exceptions.MokuException(
                f"Cannot fulfil request, error code {response.status_code}"
            )
        return response.json()

    def get_file(self, group, operation, local_path):
        url = self.url_for(group, operation)
        logger.debug(f"Downloading file from {url} to {local_path}")
        with self.rs.get(url, stream=True) as r:
            with open(local_path, "wb") as f:
                bytes_written = 0
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
                    bytes_written += len(chunk)
        logger.info(f"Downloaded {bytes_written} bytes to {local_path}")

    @handle_response
    def post_file(self, group, operation, data):
        url = self.url_for(group, operation)
        if hasattr(data, "__len__"):
            logger.debug(f"Uploading file to {url} ({len(data)} bytes)")
        else:
            logger.debug(f"Uploading file to {url} (with type {type(data)}")
        response = self.rs.post(url, data=data, timeout=self.timeout_headers())
        logger.debug(f"Upload to {url} - Status: {response.status_code}")
        return response

    @handle_response
    def delete_file(self, group, operation):
        "Deletes the given file from the Moku"
        return self.rs.delete(self.url_for(group, operation), timeout=self.timeout_headers())

    @staticmethod
    def _handle_error(code, messages):
        logger.error(f"API error: {code} - {messages}")
        if code == "NO_PLATFORM_BIT_STREAM":
            raise exceptions.NoPlatformBitstream(messages)
        elif code == "NO_BIT_STREAM":
            raise exceptions.NoInstrumentBitstream(messages)
        elif code == "INVALID_PARAM":
            raise exceptions.InvalidParameterException(messages)
        elif code == "INVALID_REQUEST":
            raise exceptions.InvalidRequestException(messages)
        elif code == "NETWORK_ERROR":
            raise exceptions.NetworkError(messages)
        elif code == "UNEXPECTED_CHANGE":
            raise exceptions.UnexpectedChangeError(messages)
        else:
            raise exceptions.MokuException(messages)

    @staticmethod
    def echo_warnings(messages):
        "Prints any warnings received from Moku"
        for m in messages or []:
            logger.warning(f"Device warning: {m}")
            print(f"Warning: {m}")

    @staticmethod
    def _normalize_nan_inf(arg):
        return {"-inf": -float("inf"), "inf": float("inf"), "nan": float("nan")}[arg]

    def _check_and_normalize_nan_inf(self, content):
        try:
            return json.loads(content)
        except json.decoder.JSONDecodeError:
            content = content.replace("nan", '"nan"')
            content = content.replace("inf", '"inf"')
            return json.loads(content, parse_constant=self._normalize_nan_inf)

    def resolve(self, response):
        "Resolves response received"

        def _parse_to_object(content):
            content = content.decode("utf-8")
            content = self._check_and_normalize_nan_inf(content)
            return namedtuple("_", content.keys())(*content.values())

        key = response.headers.get(self.sk_name)
        if key:
            self.rs.headers.update({self.sk_name: key})
        if response.status_code == 200:
            data = _parse_to_object(response.content)
            if data.success is True:
                self.echo_warnings(data.messages)
                return data.data
            elif data.success is False:
                self._handle_error(data.code, data.messages)
        else:
            # Log the full response details for debugging
            logger.debug(f"HTTP error response: {response.__dict__}")
            if response.status_code == 500:
                if response.content:
                    raise exceptions.MokuException(response.content.decode("utf-8"))
            if response.status_code == 502:
                raise exceptions.MokuException(
                    "Can't connect to the API server. Please check that your Moku has an API server installed by running `mokucli feature upload <moku> api-server`, where `<moku>` is the IP address or hardware/serial of your device, e.g. `192.168.1.123` or `MokuPro-012345`."
                )
            if response.status_code == 504:
                raise exceptions.MokuException("Timeout while waiting for a response.")
            if response.status_code == 404:
                raise exceptions.OperationNotFound(
                    "Method not found. Please make sure your Python client is compatible with the MokuOS version running on your device."
                )  # noqa
            raise exceptions.MokuException(
                f"Unknown exception. Status code: {response.status_code}"
            )



import json

from moku import Moku, MultiInstrumentSlottable
from moku.exceptions import StreamException
from moku.instruments._stream import StreamInstrument


class Datalogger(MultiInstrumentSlottable, Moku, StreamInstrument):
    """
    Datalogger instrument object.

    The Data logger instrument provides file logging of
    time-series voltage data. It contains a built-in
    Waveform Generator that can  control the analog outputs
    as well.

    Read more at https://apis.liquidinstruments.com/reference/datalogger

    """

    INSTRUMENT_ID = 7
    OPERATION_GROUP = "datalogger"

    def __init__(
        self,
        ip=None,
        serial=None,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        connect_timeout=15,
        read_timeout=30,
        slot=None,
        multi_instrument=None,
        **kwargs,
    ):
        self._init_instrument(
            ip=ip,
            serial=serial,
            force_connect=force_connect,
            ignore_busy=ignore_busy,
            persist_state=persist_state,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            slot=slot,
            multi_instrument=multi_instrument,
            **kwargs,
        )
        StreamInstrument.__init__(self, self.mokuOS_version)

    @classmethod
    def for_slot(cls, slot, multi_instrument):
        """Configures instrument at given slot in multi instrument mode"""
        return cls(slot=slot, multi_instrument=multi_instrument)

    def save_settings(self, filename):
        """
        Save instrument settings to a file. The file name should have
        a `.mokuconf` extension to be compatible with other tools.

        :type filename: FileDescriptorOrPath
        :param filename: The path to save the `.mokuconf` file to.
        """
        self.session.get_file(f"slot{self.slot}/{self.operation_group}", "save_settings", filename)

    def load_settings(self, filename):
        """
        Load a previously saved `.mokuconf` settings file into the instrument.
        To create a `.mokuconf` file, either use `save_settings` or the desktop app.

        :type filename: FileDescriptorOrPath
        :param filename: The path to the `.mokuconf` configuration to load
        """
        with open(filename, "rb") as f:
            self.session.post_file(
                f"slot{self.slot}/{self.operation_group}", "load_settings", data=f
            )

    def set_frontend(self, channel, impedance, coupling, range, bandwidth=None, strict=True):
        """
        set_frontend.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type impedance: `string` ['1MOhm', '50Ohm']
        :param impedance: Input impedance

        :type coupling: `string` ['AC', 'DC']
        :param coupling: Input coupling

        :type range: `string` ['100mVpp', '400mVpp', '1Vpp', '2Vpp', '4Vpp', '10Vpp', '40Vpp', '50Vpp'] # noqa
        :param range: Input range

        :type bandwidth: `string` ['1MHz', '30MHz', '200MHz', '300MHz', '600MHz', '2GHz']
        :param bandwidth: Input bandwidth

        """
        operation = "set_frontend"
        params = dict(
            strict=strict,
            channel=channel,
            impedance=impedance,
            coupling=coupling,
            range=range,
            bandwidth=bandwidth,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_frontend(self, channel):
        """
        get_frontend.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_frontend"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_acquisition_mode(self, mode="Normal", strict=True):
        """
        set_acquisition_mode.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type mode: `string` ['Normal', 'Precision', 'DeepMemory', 'PeakDetect'] # noqa
        :param mode: Acquisition Mode

        """
        operation = "set_acquisition_mode"
        params = dict(strict=strict, mode=mode)
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_acquisition_mode(self):
        """
        get_acquisition_mode.
        """
        operation = "get_acquisition_mode"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_samplerate(self, sample_rate, strict=True):
        """
        set_samplerate.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type sample_rate: `number` [10, 1e6]
        :param sample_rate: Target samples per second

        """
        operation = "set_samplerate"
        params = dict(
            strict=strict,
            sample_rate=sample_rate,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_samplerate(self):
        """
        get_samplerate.
        """
        operation = "get_samplerate"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def disable_channel(self, channel, disable=True, strict=True):
        """
        .. deprecated:: 3.1.1
        Use `enable_input` instead.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type disable: `boolean`
        :param disable: Boolean value to enable/disable

        """
        operation = "disable_channel"
        params = dict(
            strict=strict,
            channel=channel,
            disable=disable,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def enable_input(self, channel, enable=True, strict=True):
        """
        enable_input.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type enable: `boolean`
        :param enable: Enable input signal

        """
        operation = "enable_input"
        params = dict(
            strict=strict,
            channel=channel,
            enable=enable,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def start_logging(
        self,
        duration=60,
        delay=0,
        file_name_prefix="",
        comments="",
        trigger_source=None,
        trigger_level=None,
        strict=True,
        sample_rate=None,
    ):
        """
        start_logging.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type duration: `double` Sec (defaults to 60)
        :param duration: Duration to log for

        :type delay: `integer` Sec (defaults to 0)
        :param delay: Delay the start by

        :type trigger_source: `string` ['ChannelA', 'ChannelB', 'ChannelC', 'ChannelD', 'Input1', 'Input2', 'Input3', 'Input4', 'Output1', 'Output2', 'Output3', 'Output4', 'External'] # noqa
        :param trigger_source: Trigger source

        :type trigger_level: `number` [-5V, 5V]  (defaults to 0)
        :param trigger_level: Trigger level

        :type file_name_prefix: `string`
        :param file_name_prefix: Optional file name prefix

        :type comments: `string`
        :param comments: Optional comments to be included


        .. important::
            It is recommended **not** to relinquish the ownership of the
            device until logging session is completed

        """
        operation = "start_logging"
        params = dict(
            strict=strict,
            duration=duration,
            sample_rate=sample_rate,
            delay=delay,
            trigger_source=trigger_source,
            trigger_level=trigger_level,
            file_name_prefix=file_name_prefix,
            comments=comments,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def stop_logging(self):
        """
        stop_logging.
        """
        operation = "stop_logging"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def summary(self):
        """
        summary.
        """
        operation = "summary"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def generate_waveform(
        self,
        channel,
        type,
        amplitude=1,
        frequency=10000,
        offset=0,
        phase=0,
        duty=None,
        symmetry=None,
        dc_level=None,
        edge_time=None,
        pulse_width=None,
        strict=True,
    ):
        """
        generate_waveform.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type type: `string` ['Off', 'Sine', 'Square', 'Ramp', 'Pulse', 'Noise', 'DC'] # noqa
        :param type: Waveform type

        :type amplitude: `number` [4e-3V, 10V]  (defaults to 1)
        :param amplitude: Waveform peak-to-peak amplitude

        :type frequency: `number` [1e-3Hz, 20e6Hz]  (defaults to 10000)
        :param frequency: Waveform frequency

        :type offset: `number` [-5V, 5V]  (defaults to 0)
        :param offset: DC offset applied to the waveform

        :type phase: `number` [0Deg, 360Deg]  (defaults to 0)
        :param phase: Waveform phase offset

        :type duty: `number` [0.0%, 100.0%]
        :param duty: Duty cycle as percentage (Only for Square wave)

        :type symmetry: `number` [0.0%, 100.0%]
        :param symmetry: Fraction of the cycle rising

        :type dc_level: `number`
        :param dc_level: DC Level. (Only for DC waveform)

        :type edge_time: `number` [16e-9, pulse width]
        :param edge_time: Edge-time of the waveform (Only for Pulse wave)

        :type pulse_width: `number`
        :param pulse_width: Pulse width of the waveform (Only for Pulse wave)

        """
        operation = "generate_waveform"
        params = dict(
            strict=strict,
            channel=channel,
            type=type,
            amplitude=amplitude,
            frequency=frequency,
            offset=offset,
            phase=phase,
            duty=duty,
            symmetry=symmetry,
            dc_level=dc_level,
            edge_time=edge_time,
            pulse_width=pulse_width,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def sync_output_phase(self):
        """
        sync_output_phase.
        """
        operation = "sync_output_phase"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def set_output_load(self, channel, load, strict=True):
        """
        .. deprecated:: 3.1.1
        Use `set_output_termination` instead.

        set_output_load.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type load: `string` ['1MOhm', '50Ohm']
        :param load: Output load

        """
        operation = "set_output_load"
        params = dict(
            strict=strict,
            channel=channel,
            load=load,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output_load(self, channel):
        """
        .. deprecated:: 3.1.1
        Use `get_output_termination` instead.

        get_output_load.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output_load"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_output_termination(self, channel, termination, strict=True):
        """
        set_output_termination.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type termination: `string` ['HiZ', '50Ohm']
        :param termination: Output termination

        """
        operation = "set_output_termination"
        params = dict(
            strict=strict,
            channel=channel,
            termination=termination,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output_termination(self, channel):
        """
        get_output_termination.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output_termination"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def logging_progress(self):
        """
        logging_progress.
        """
        operation = "logging_progress"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def start_streaming(
        self, duration=None, sample_rate=None, trigger_source=None, trigger_level=None
    ):
        """
        start_streaming.

        :type duration: `double`
        :param duration: Duration in second(s) to stream for

        :type sample_rate: `number` [10, 1e6]
        :param sample_rate: Target samples per second

        :type trigger_source: `string` ['ChannelA', 'ChannelB', 'ChannelC', 'ChannelD', 'Input1', 'Input2', 'Input3', 'Input4', 'Output1', 'Output2', 'Output3', 'Output4', 'External'] # noqa
        :param trigger_source: Trigger source

        :type trigger_level: `number` [-5V, 5V]  (defaults to 0)
        :param trigger_level: Trigger level

        """
        super().start_streaming()
        operation = "start_streaming"
        params = dict(
            duration=duration,
            sample_rate=sample_rate,
            trigger_source=trigger_source,
            trigger_level=trigger_level,
        )
        response = self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)
        self.stream_id = response["stream_id"]
        self.ip_address = self.session.ip_address
        return response

    def stop_streaming(self):
        """
        stop_streaming.

        """
        operation = "stop_streaming"
        response = self.session.post(f"slot{self.slot}/{self.operation_group}", operation)
        self.stream_id = None
        return response

    def get_chunk(self):
        """
        get_chunk.

        Get the next raw chunk from the streaming session

        """
        data = {"stream_id": {self.stream_id: {"topic": f"logformat{self.slot - 1}"}}}
        result = self.session.post_to_v2_raw("get_chunk", params=data)

        if result.status_code != 200:
            raise StreamException("Error fetching stream.")

        try:
            error = json.loads(result.content)
            raise StreamException(error.get("error", error))
        except StreamException as e:
            raise e
        except Exception:
            return result.content

    def get_stream_status(self):
        """
        get_stream_status.

        """
        operation = "get_stream_status"

        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)



import tarfile
import tempfile
from pathlib import Path

from moku import Moku, MultiInstrumentSlottable
from moku.exceptions import MokuException, NoInstrumentBitstream


class CustomInstrument(MultiInstrumentSlottable, Moku):
    """
    Custom Instrument

    """

    INSTRUMENT_ID = 255
    OPERATION_GROUP = "custominstrument"

    def __init__(
        self,
        ip=None,
        serial=None,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        bitstream=None,
        connect_timeout=15,
        read_timeout=30,
        slot=None,
        multi_instrument=None,
        **kwargs,
    ):
        assert bitstream, "Bitstream package path is required for Custom Instrument"

        bitstream_path = Path(bitstream)
        if not bitstream_path.exists():
            raise FileNotFoundError(f"Bitstream package not found at {bitstream_path}")

        with tempfile.TemporaryDirectory(prefix="moku_cloudcompile_") as temp_dir:
            # Open and extract the tar/tar.gz file
            with tarfile.open(bitstream_path, mode="r:*") as tar:
                tar.extractall(temp_dir)

            try:
                self._init_instrument(
                    ip=ip,
                    serial=serial,
                    force_connect=force_connect,
                    ignore_busy=ignore_busy,
                    persist_state=persist_state,
                    connect_timeout=connect_timeout,
                    read_timeout=read_timeout,
                    slot=slot,
                    multi_instrument=multi_instrument,
                    bs_path=temp_dir,
                    **kwargs,
                )
            except NoInstrumentBitstream:
                # Intercept this exception to change the error message. By default this will show the
                # "run mokucli instrument download" message, which is not applicable here.
                raise MokuException(
                    f"Failed to initialize custom instrument. Please check the bitstream package and try again (tried {bitstream_path.absolute()})"
                )

    @classmethod
    def for_slot(cls, slot, multi_instrument, **kwargs):
        """Configures instrument at given slot in multi instrument mode"""
        bitstream = kwargs.get("bitstream")
        return cls(slot=slot, multi_instrument=multi_instrument, bitstream=bitstream)

    def save_settings(self, filename):
        """
        Save instrument settings to a file. The file name should have
        a `.mokuconf` extension to be compatible with other tools.

        :type filename: FileDescriptorOrPath
        :param filename: The path to save the `.mokuconf` file to.
        """
        self.session.get_file(f"slot{self.slot}/{self.operation_group}", "save_settings", filename)

    def load_settings(self, filename):
        """
        Load a previously saved `.mokuconf` settings file into the instrument.
        To create a `.mokuconf` file, either use `save_settings` or the desktop app.

        :type filename: FileDescriptorOrPath
        :param filename: The path to the `.mokuconf` configuration to load
        """
        with open(filename, "rb") as f:
            self.session.post_file(
                f"slot{self.slot}/{self.operation_group}", "load_settings", data=f
            )

    def set_controls(self, controls, strict=True):
        """
        set_controls.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type controls: `list`
        :param controls: List of control map(pair of id, value)

        """
        operation = "set_controls"
        params = dict(
            strict=strict,
            controls=controls,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_control(self, idx, value, strict=True):
        """
        set_control.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type idx: `integer`
        :param idx: Control ID(0 indexed)

        :type value: `integer`
        :param value: Register value

        """
        operation = "set_control"
        params = dict(
            strict=strict,
            idx=idx,
            value=value,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_control(self, idx, strict=True):
        """
        get_control.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type idx: `integer`
        :param idx: Control ID(0 indexed)

        """
        operation = "get_control"
        params = dict(
            strict=strict,
            idx=idx,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_controls(self):
        """
        get_controls.
        """
        operation = "get_controls"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def get_status(self):
        """
        get_status.
        """
        operation = "get_status"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_interpolation(self, channel, enable=True, strict=True):
        """
        set_interpolation.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type enable: `boolean`
        :param enable: Enable/disable interpolation on specified channel

        """
        operation = "set_interpolation"
        params = dict(
            strict=strict,
            channel=channel,
            enable=enable,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def sync(self, mask, strict=True):
        """
        sync.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type mask: `integer`
        :param mask: Mask value

        """
        operation = "sync"
        params = dict(
            strict=strict,
            mask=mask,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_interpolation(self, channel):
        """
        get_interpolation.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_interpolation"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def summary(self):
        """
        summary.
        """
        operation = "summary"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)


class CustomInstrumentPlus(CustomInstrument):
    """
    CustomInstrumentPlus

    """

    INSTRUMENT_ID = 254
    OPERATION_GROUP = "custominstrumentplus"


class CloudCompile(CustomInstrument):
    def __init__(self, *args, **kwargs):
        # deprecated in 4.1.1
        print("Warning: CloudCompile is deprecated, use CustomInstrument instead")
        super().__init__(*args, **kwargs)



from moku import Moku, MultiInstrumentSlottable


class GigabitStreamer(MultiInstrumentSlottable, Moku):
    """
    GigabitStreamer instrument object.

    This instrument streams data by transmitting and/or receiving UDP packets through
    the SFP ports at gigabit speeds.

    Read more at https://apis.liquidinstruments.com/reference/gs

    """

    INSTRUMENT_ID = 12
    OPERATION_GROUP = "gs"

    def __init__(
        self,
        ip=None,
        serial=None,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        connect_timeout=15,
        read_timeout=30,
        slot=None,
        multi_instrument=None,
        **kwargs,
    ):
        self._init_instrument(
            ip=ip,
            serial=serial,
            force_connect=force_connect,
            ignore_busy=ignore_busy,
            persist_state=persist_state,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            slot=slot,
            multi_instrument=multi_instrument,
            **kwargs,
        )

    @classmethod
    def for_slot(cls, slot, multi_instrument):
        """Configures instrument at given slot in multi instrument mode"""
        return cls(slot=slot, multi_instrument=multi_instrument)

    def save_settings(self, filename):
        """
        Save instrument settings to a file. The file name should have
        a `.mokuconf` extension to be compatible with other tools.

        :type filename: FileDescriptorOrPath
        :param filename: The path to save the `.mokuconf` file to.
        """
        self.session.get_file(f"slot{self.slot}/{self.operation_group}", "save_settings", filename)

    def load_settings(self, filename):
        """
        Load a previously saved `.mokuconf` settings file into the instrument.
        To create a `.mokuconf` file, either use `save_settings` or the desktop app.

        :type filename: FileDescriptorOrPath
        :param filename: The path to the `.mokuconf` configuration to load
        """
        with open(filename, "rb") as f:
            self.session.post_file(
                f"slot{self.slot}/{self.operation_group}", "load_settings", data=f
            )

    def summary(self):
        """
        summary.
        """
        operation = "summary"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_defaults(self):
        """
        set_defaults.
        """
        operation = "set_defaults"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def set_frontend(
        self,
        channel,
        impedance,
        coupling,
        attenuation=None,
        gain=None,
        bandwidth=None,
        strict=True,
    ):
        """
        set_frontend.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type impedance: `string` ['1MOhm', '50Ohm']
        :param impedance: Input impedance

        :type coupling: `string` ['AC', 'DC']
        :param coupling: Input coupling

        :type attenuation: `string` ['-20dB', '0dB', '14dB', '20dB', '32dB', '40dB'] # noqa
        :param attenuation: Input attenuation.

        :type gain: `string` ['20dB', '0dB', '-14dB', '-20dB', '-32dB', '-40dB'] # noqa
        :param gain: Input gain.

        :type bandwidth: `string` ['1MHz', '30MHz', '200MHz', '300MHz', '600MHz', '2GHz']
        :param bandwidth: Input bandwidth

        """
        operation = "set_frontend"
        params = dict(
            strict=strict,
            channel=channel,
            impedance=impedance,
            coupling=coupling,
            attenuation=attenuation,
            gain=gain,
            bandwidth=bandwidth,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_frontend(self, channel):
        """
        get_frontend.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_frontend"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def enable_input(self, channel, enable=True, strict=True):
        """
        enable_input.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type enable: `boolean`
        :param enable: Enable input signal

        """
        operation = "enable_input"
        params = dict(
            strict=strict,
            channel=channel,
            enable=enable,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_acquisition(self, mode, sample_rate, strict=True):
        """
        set_acquisition.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type mode: `string` ['Normal', 'Precision'] # noqa
        :param mode: Acquisition mode

        :type sample_rate: `number` [5e3, 5e9]
        :param sample_rate: Target samples per second

        """
        operation = "set_acquisition"
        params = dict(
            strict=strict,
            mode=mode,
            sample_rate=sample_rate,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def start_sending(self, duration, delay=0, trigger_source=None, trigger_level=0, strict=True):
        """
        start_sending.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type duration: `double` Sec
        :param duration: Duration to log for

        :type delay: `integer` Sec
        :param delay: Delay the start by

        :type trigger_source: `string` ['Input1', 'Input2', 'Input3', 'Input4', 'InputA', 'InputB', 'External'] # noqa
        :param trigger_source: Trigger source

        :type trigger_level: `number` [-5V, 5V]
        :param trigger_level: Trigger level

        """
        operation = "start_sending"
        params = dict(
            strict=strict,
            duration=duration,
            delay=delay,
            trigger_source=trigger_source,
            trigger_level=trigger_level,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def stop_sending(self):
        """
        stop_sending.
        """
        operation = "stop_sending"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def get_send_status(self):
        """
        get_send_status.
        """
        operation = "get_send_status"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def get_receive_status(self):
        """
        get_receive_status.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        """
        operation = "get_receive_status"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def reset_receive_counters(self):
        """
        reset_receive_counters.
        """
        operation = "reset_receive_counters"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def set_local_network(self, ip_address, port, multicast_ip_address=None, strict=True):
        """
        set_local_network.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type ip_address: `string`
        :param ip_address: IP address

        :type port: `integer`
        :param port: UDP port number

        :type multicast_ip_address: `string`
        :param multicast_ip_address: Multicast IP address

        """
        operation = "set_local_network"
        params = dict(
            strict=strict,
            ip_address=ip_address,
            multicast_ip_address=multicast_ip_address,
            port=port,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_outgoing_packets(self, mtu, strict=True):
        """
        set_outgoing_packets.

        :type mtu: `string` ["508bytes", "576bytes", "1500bytes", "9000bytes", "65535bytes"] # noqa
        :param mtu: Network Maximum Transmission Unit

        """
        operation = "set_outgoing_packets"
        params = dict(
            strict=strict,
            mtu=mtu,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_remote_network(self, ip_address, port, mac_address, strict=True):
        """
        set_remote_network.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type ip_address: `string`
        :param ip_address: IP Address

        :type port: `integer`
        :param port: Remote UDP port number

        :type mac_address: `string`
        :param mac_address: MAC address

        """
        operation = "set_remote_network"
        params = dict(
            strict=strict,
            ip_address=ip_address,
            port=port,
            mac_address=mac_address,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_interpolation(self, mode, strict=True):
        """
        set_interpolation.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type mode: `string` ["None", "Linear"]
        :param mode: Interpolation mode

        """
        operation = "set_interpolation"
        params = dict(
            strict=strict,
            mode=mode,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_output(self, channel, enable, gain, offset, strict=True):
        """
        set_output.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type enable: `boolean`
        :param enable: Enable output signal

        :type gain: `number`
        :param gain: Gain in dB

        :type offset: `number`
        :param offset: Offset in V

        """
        operation = "set_output"
        params = dict(
            strict=strict,
            channel=channel,
            enable=enable,
            gain=gain,
            offset=offset,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)


class GigabitStreamerPlus(GigabitStreamer):
    """
    GigabitStreamerPlus instrument object.

    This instrument streams data by transmitting and/or receiving UDP packets through
    the QSFP port at high gigabit speeds.

    Read more at https://apis.liquidinstruments.com/reference/gsp

    """

    INSTRUMENT_ID = 13
    OPERATION_GROUP = "gsp"



"""

Author: Aaron - Aqacia Pty Ltd for Liquid Instruments
        Alex Mason - Liquid Instruments
Date: 2024
"""

import json
import sys
import numpy as np
import tensorflow as tf
import tensorflow.keras as keras
import logging
from typing import Any, Optional, List, Tuple
from copy import copy, deepcopy

# list of supported activations
_AVAILABLE_ACTIVATIONS = ["relu", "tanh", "sigmoid", "softsign", "linear"]
_LOG = logging.getLogger("Linn")


def list_activations():
    """
    List the available activation functions
    :return: List of available activation functions
    """
    return _AVAILABLE_ACTIVATIONS


class WeightClip(keras.constraints.Constraint):
    def __init__(self, clip_value: float = 1.0) -> None:
        """
        Clips all the weights/biases to be within the range [-1, 1] to help with quantization later on.
        :param clip_value: value to clip the model between (float).
        """
        super().__init__()
        self.clip_value = float(clip_value)

    def __call__(self, w: Any):
        """
        Clip the weights/biases to the symmetric bounds
        :param w: Tensor or variable representing the weights
        :return: Tensor or variable clipped to the bounds
        """
        return tf.clip_by_value(w, -self.clip_value, self.clip_value)

    def get_config(self):
        return {"name": self.__class__.__name__, "clip_value": self.clip_value}


class OutputClipLayer(keras.layers.Layer):
    def __init__(self, clip_value: float = 1.0):
        """
        Class for clipping the output layers of the network. Ensures that outputs are clipped to [-1, 1] bounds for
        quantization.
        """
        super().__init__()
        self.clip_value = float(clip_value)

    def call(self, inputs, *args, **kwargs):
        """
        Clip the inputs to the symmetric bounds
        :param inputs: The output of the previous layer
        :param args: Mirrored from base layer class.
        :param kwargs: Mirrored from base layer class.
        :return: Tensor or variable clipped to the bounds
        """
        return tf.clip_by_value(inputs, -self.clip_value, self.clip_value)


class LinnModel:
    def __init__(self):
        self.model = None
        self.training_inputs = None
        self.training_outputs = None

        self._input_transform_args = None
        self._output_transform_args = None

        self.model_definition = None

    def construct_model(
        self,
        layer_definition: list = None,
        show_summary: bool = False,
        optimizer: any = "adam",
        loss: any = "mse",
        metrics: any = (),
    ) -> None:
        """
        Construct the model to be used by the rest of the functions in this class.
        :param layer_definition: a list of tuples of the form `(layer_width, activation)`
        which defines the model. If not provided the default model will be used.
        `(layer_width,)` can be used to signify a linear (identity) activation function.
        :param show_summary: (bool) determines whether the model summary is displayed
        :param optimizer: optimizer fed to the keras compile option.
        :param loss: loss function fed to the keras compile option.
        :param metrics: metrics for the model to track during training
        :return: None
        """

        # check the input and output
        if self.training_inputs is None or self.training_outputs is None:
            raise TypeError("Please set the training data first.")

        # check the model definition type
        if layer_definition is not None:
            if type(layer_definition) is not list:
                raise TypeError(
                    f"Expected type:<list> for model definition. Received: <{type(layer_definition)}>."
                )
            elif len(layer_definition) == 0:
                raise ValueError("layer_definition can't be empty.")
        else:
            raise ValueError("layer_definition can't be empty.")

        # assign the definition for checking
        self.model_definition = layer_definition

        # construct the model
        input_layer = keras.layers.Input((self.training_inputs.shape[-1],))
        prev_layer = input_layer
        # prev_layer = None
        for idx, layer_defn in enumerate(layer_definition):
            if type(layer_defn) is not tuple:
                raise TypeError(
                    f"layer_definition index:{idx} is type:<{type(layer_defn)} not <tuple>."
                )

            if len(layer_defn) != 2:
                raise ValueError(
                    f"layer_definition index:{idx} should have length 2, not {len(layer_defn)}."
                )

            # grab the activation function
            # activation = 'linear'

            activation = str(layer_defn[1])
            if activation not in list_activations():
                raise ValueError(
                    f"Activation {activation} is not in the list of supported activations. "
                    f"Try {list_activations()}."
                )
            # grab the layer width
            try:
                layer_width = int(layer_defn[0])
            except ValueError as e:
                raise ValueError(f"Error converting the layer width: {e.args}.")

            prev_layer = keras.layers.Dense(
                layer_width,
                activation=activation,
                kernel_constraint=WeightClip(),
                bias_constraint=WeightClip(),
            )(prev_layer)
            prev_layer = OutputClipLayer()(prev_layer)

        # compile the model for training
        self.model = keras.Model(input_layer, prev_layer)
        self.model.compile(optimizer=optimizer, loss=loss, metrics=[met for met in metrics])

        # summarize if necessary
        if show_summary:
            self.model.summary()

    def _check_model_definition(self):
        # Check the input_layer first
        if type(self.model_definition[0]) is not tuple:
            _LOG.warning(
                f"Definition of input_layer is type:<{type(self.model_definition[0])} not <tuple>."
            )

        if len(self.model_definition[0]) != 1:
            _LOG.warning(
                f"Definition of input_layer should have length 1, not {len(self.model_definition[0])}."
            )

        if len(self.model_definition[0]) == 2:
            _LOG.warning("Ignored the activation function in the input_layer")

        if self.model_definition[0][0] != self.training_inputs.shape[-1]:
            _LOG.warning(
                f"Shape of layer_definition index: 0 not match the shape of input data, should be "
                f"{self.training_inputs.shape[-1]}, not {self.model_definition[0][0]}. Changed to the shape of"
                f" input data automatically."
            )

        self.model_definition[0] = (self.training_inputs.shape[-1],)

        for idx, layer_defn in enumerate(self.model_definition[1:]):
            if type(layer_defn) is not tuple:
                raise TypeError(
                    f"layer_definition index:{idx + 1} is type:<{type(layer_defn)} not <tuple>."
                )

            if len(layer_defn) not in [1, 2]:
                raise ValueError(
                    f"layer_definition index:{idx + 1} should have length 1 or 2, not {len(layer_defn)}."
                )

            if len(layer_defn) == 2:
                if layer_defn[1] not in list_activations():
                    raise ValueError(
                        f"Activation {layer_defn[1]} is not in the list of supported activations. "
                        f"Try {list_activations()}."
                    )

        if self.model_definition[-1][0] != self.training_outputs.shape[-1]:
            _LOG.warning(
                f"Shape of layer_definition index: {len(self.model_definition) - 1} does not match the shape of "
                f"output data, should be {self.training_outputs.shape[1:]}, not "
                f"{self.model_definition[-1][0]}. Changed to the shape of output data automatically"
            )
            if len(self.model_definition[-1]) == 2:
                self.model_definition[-1] = (
                    self.training_outputs.shape[-1],
                    self.model_definition[-1][1],
                )
            else:
                self.model_definition[-1] = (self.training_outputs.shape[-1],)

        return True

    def _check_data_model_dim(self, inputs: np.ndarray, outputs: np.ndarray, data_name: str):
        """
        Check the data and model dimensions against one another to make sure they are correct.
        :param inputs: Inputs to the model
        :param outputs: Outputs of the model
        :param data_name: String indicating whether it's training or validation.
        :return: Boolean indicating whether everything is ok
        """
        # warn the user dimension checking is skipped
        if self.model is None:
            _LOG.warning("Keras model is not initialized. Dimension checking will be skipped.")
            return False

        # check if the dimensions match the defined model
        if (
            len(inputs.shape) != len(self.model.input_shape)
            or inputs.shape[1:] != self.model.input_shape[1:]
        ):
            raise ValueError(
                f"Dimensions of {data_name} inputs do not match model: {inputs.shape[1:]}, "
                f"{self.model.input_shape[1:]}"
            )

        if (
            len(outputs.shape) != len(self.model.output_shape)
            or outputs.shape[1:] != self.model.output_shape[1:]
        ):
            raise ValueError(
                f"Dimensions of {data_name} outputs do not match model: {outputs.shape[1:]}, "
                f"{self.model.output_shape[1:]}"
            )

        # if we got this far, success!
        return True

    def _transform(self, data_array: np.ndarray, is_input: bool):
        """
        Transform the data into a domain amenable for quantization.
        :param data_array: Data array to be transformed
        :param is_input: is it an input or output?
        :return: Return the transformed array
        """
        # copy the array so we don't modify the original
        output_array = np.copy(data_array)
        # get the scales if they are missing
        if is_input and self._input_transform_args is None:
            self._input_transform_args = (
                np.min(output_array, axis=0),
                np.max(output_array, axis=0),
            )
        if not is_input and self._output_transform_args is None:
            self._output_transform_args = (
                np.min(output_array, axis=0),
                np.max(output_array, axis=0),
            )

        # scale everything according to the transform arguments
        if is_input:
            output_array -= self._input_transform_args[0]
            output_array /= self._input_transform_args[1] - self._input_transform_args[0]
        else:
            output_array -= self._output_transform_args[0]
            output_array /= self._output_transform_args[1] - self._output_transform_args[0]
        output_array = 2 * (output_array - 0.5)

        return output_array

    def _inverse_transform(self, data_array: np.ndarray, is_input: bool):
        """
        Undo the scaling of the transform function to bring data back to the original domain.
        :param data_array: Data to be transformed
        :param is_input: is it the input or output
        :return: Return the transformed array
        """
        # copy the array so we don't destroy the original
        output_array = np.copy(data_array)
        # see if things can be transformed
        if is_input and self._input_transform_args is None:
            raise RuntimeError("Cannot inverse transform before first calling transform.")
        if not is_input and self._output_transform_args is None:
            raise RuntimeError("Cannot inverse transform before first calling transform.")

        # perform the inverse scaling
        output_array /= 2.0
        output_array += 0.5
        if is_input:
            output_array *= self._input_transform_args[1] - self._input_transform_args[0]
            output_array += self._input_transform_args[0]
        else:
            output_array *= self._output_transform_args[1] - self._output_transform_args[0]
            output_array += self._output_transform_args[0]

        return output_array

    def set_training_data(
        self,
        training_inputs: np.ndarray,
        training_outputs: np.ndarray,
        scale: bool = True,
        input_data_boundary: Tuple[list] = None,
        output_data_boundary: Tuple[list] = None,
    ):
        """
        Set the training data to be used by the model and automatically scale them to use the full dynamic range if
        specified. Scaling will ensure the model inputs and outputs are in the domain [-1, 1]. Call this before
        constructing a model to assign the data dimensions in construction.
        :param training_inputs: numpy.ndarray of input training data which should match the model dimensions
        :param training_outputs: numpy.ndarray of output training data which should match the model dimensions
        :param scale: (bool) automatically scale the data to fit in the bounds.
        :param input_data_boundary: tuple of the input data boundaries that should be used for scaling
        :param output_data_boundary: tuple of the output data boundaries that should be used for scaling
        :return: None
        """

        # We copy the input data so the scaling doesn't affect the original. This can be big, so might
        # want to revisit and e.g. only copy if we're scaling
        training_inputs = np.array(copy(training_inputs), dtype=float)
        training_outputs = np.array(copy(training_outputs), dtype=float)

        if not scale and (input_data_boundary is not None or output_data_boundary is not None):
            _LOG.warning("scale = False. Ignoring the boundary setting")

        if scale:
            if input_data_boundary is not None:
                if (
                    not isinstance(input_data_boundary, (list, tuple))
                    or len(input_data_boundary) != 2
                ):
                    raise TypeError(
                        "Expected tuple of the form (lower_bound, upper_bound) for boundary setting."
                    )

                if not isinstance(input_data_boundary[0], (np.ndarray, list)):
                    raise TypeError(
                        f"Lower bounds should be type:<numpy.ndarray> or type:<list> not {type(input_data_boundary[0])}."
                    )

                if not isinstance(input_data_boundary[1], (np.ndarray, list)):
                    raise TypeError(
                        f"Upper bounds should be type:<numpy.ndarray> or type:<list> not {type(input_data_boundary[1])}."
                    )

                if len(input_data_boundary[0]) != training_inputs.shape[-1]:
                    raise ValueError(
                        f"Dimensions of lower bounds do not match input: {len(input_data_boundary[0])}, "
                        f"{training_inputs.shape[-1]}"
                    )
                if len(input_data_boundary[1]) != training_inputs.shape[-1]:
                    raise ValueError(
                        f"Dimensions of upper bounds do not match input: {len(input_data_boundary[1])}, "
                        f"{training_inputs.shape[-1]}"
                    )

            if output_data_boundary is not None:
                if (
                    not isinstance(output_data_boundary, (list, tuple))
                    or len(output_data_boundary) != 2
                ):
                    raise TypeError(
                        "Expected tuple of the form (lower_bound, upper_bound) for boundary setting."
                    )

                if not isinstance(output_data_boundary[0], (np.ndarray, list)):
                    raise TypeError(
                        f"Lower bounds should be type:<numpy.ndarray> or type:<list> not {type(output_data_boundary[0])}."
                    )

                if not isinstance(output_data_boundary[1], (np.ndarray, list)):
                    raise TypeError(
                        f"Upper bounds should be type:<numpy.ndarray> or type:<list> not {type(output_data_boundary[1])}."
                    )

                if len(output_data_boundary[0]) != training_outputs.shape[-1]:
                    raise ValueError(
                        f"Dimensions of lower bounds do not match outputs: {len(input_data_boundary[0])}, "
                        f"{training_outputs.shape[-1]}"
                    )
                if len(output_data_boundary[1]) != training_outputs.shape[-1]:
                    raise ValueError(
                        f"Dimensions of upper bounds do not match outputs: {len(output_data_boundary[1])}, "
                        f"{training_outputs.shape[-1]}"
                    )

        if input_data_boundary is not None:
            self._input_transform_args = [np.array(boundary) for boundary in input_data_boundary]
        if output_data_boundary is not None:
            self._output_transform_args = [np.array(boundary) for boundary in output_data_boundary]

        # scale the data if necessary to scale the full dynamic range
        if scale:
            training_inputs = self._transform(training_inputs, True)
            training_outputs = self._transform(training_outputs, False)

        # set the internal variables
        self.training_inputs = training_inputs
        self.training_outputs = training_outputs

    @staticmethod
    def _log_missing_value(key: str, default_value: Any, config: dict):
        """
        Convenience function for informing the user they have failed to provide a value and the defaul will be used.
        :param key: Key associated with the lookup.
        :param default_value: default value that is being used.
        :param config: dictionary to be read from
        :return: value corresponding to the default or dictionary value
        """
        dict_value = config.get(key, None)
        if dict_value is None:
            _LOG.warning(f"Value for {key} missing. Using default:{default_value}.")
            return default_value
        else:
            return dict_value

    def fit_model(
        self,
        epochs: int,
        es_config: Optional[dict] = None,
        validation_split: float = 0.0,
        validation_data: Optional[tuple] = None,
        **keras_kwargs: Optional[Any],
    ):
        """
        Fit the model according to the data that is stored in the training inputs and outputs.
        set_training_data and construct_model must be called prior to calling Model.fit_model()
        :param epochs: (int) number of epochs to train for
        :param es_config: configuration dictionary for the early stopping callback
        :param validation_split: (float) used to define the validation split
        :param validation_data: validation data in a tuple of form (inputs, outputs)
        :param keras_kwargs: keyword args to pass to the keras `fit` function
        :return: history object from the keras `fit` function
        """

        # cast to an integer or throw an error
        epochs = int(epochs)

        validation_split = float(validation_split)
        if validation_split < 0.0 or validation_split > 1.0:
            _LOG.warning("Validation split outside of bounds - coercing value.")
            validation_split = np.clip(validation_split, 0.0, 1.0)

        # check the validation data
        if validation_data is not None:
            if type(validation_data) is not tuple or len(validation_data) != 2:
                raise TypeError(
                    "Expected tuple of the form (val_input, val_output) for validation data."
                )

            if type(validation_data[0]) is not np.ndarray:
                raise TypeError(
                    f"Validation input should be type:<numpy.ndarray> not <{type(validation_data[0])}>."
                )

            if type(validation_data[1]) is not np.ndarray:
                raise TypeError(
                    f"Validation output should be type:<numpy.ndarray> not <{type(validation_data[1])}>."
                )

            # check the data dimensions
            success = self._check_data_model_dim(
                validation_data[0], validation_data[1], "validation"
            )
            if not success:
                raise RuntimeError(
                    f"Model is not initialized. Call {self.__class__.__name__}.construct_model to create"
                    f" the model."
                )

        # warn of a redundancy
        if validation_split > 0.0 and (validation_data is not None):
            _LOG.warning("Supplied validation_data overrides validation_split.")

        # callbacks to use
        callbacks = []

        # build the early stopping call back
        if es_config is not None:
            # make sure it's a dictionary
            if type(es_config) is not dict:
                raise TypeError(
                    f"Expected type:<dict> for early stopping configuration, not <{type(es_config)}>."
                )

            # parameters for the callback
            monitor = self._log_missing_value("monitor", "val_loss", es_config)
            patience = self._log_missing_value("patience", 5, es_config)
            restore = self._log_missing_value("restore", False, es_config)

            # callback to be used later
            es_callback = keras.callbacks.EarlyStopping(
                monitor=monitor,
                min_delta=0,
                patience=patience,
                verbose=0,
                mode="auto",
                baseline=None,
                restore_best_weights=restore,
            )
            callbacks.append(es_callback)

            # fail if there is no validation, yet we monitor it.
            if validation_split == 0.0 and validation_data is None and monitor == "val_loss":
                raise RuntimeError(
                    "Cannot monitor val_loss for early stopping with val_split=0.0 and val_data=None."
                )

        # finally fit the model
        if validation_data is not None:
            history = self.model.fit(
                self.training_inputs,
                self.training_outputs,
                epochs=epochs,
                validation_data=validation_data,
                callbacks=callbacks,
                **keras_kwargs,
            )
        else:
            history = self.model.fit(
                self.training_inputs,
                self.training_outputs,
                epochs=epochs,
                validation_split=validation_split,
                callbacks=callbacks,
                **keras_kwargs,
            )

        return history

    def predict(
        self,
        inputs: np.ndarray,
        scale: Optional[bool] = None,
        unscale_output: Optional[bool] = None,
        **keras_kwargs: Optional[Any],
    ):
        """
        Generates model predictions for given inputs, with optional scaling based on training settings.
        :param inputs: The input data to be run through the model
        :param scale: Whether the data should be scaled at the input. Scaling will ensure the model inputs are
        scaled as set in set_training_data and `scale = None` will scale data if trained with scaling.
        :param unscale_output: Whether the data should be unscaled at the output. Unscaling will ensure the model
        outputs are scaled as set in set_training_data and `unscale = None` will unscale data if trained with unscaling.
        :param keras_kwargs: parameters to be passed to the keras predict function if needed.
        :return: The model predictions
        """

        # Default input and output scaling to whatever was used in training
        if scale is None:
            scale = self._input_transform_args is not None

        if unscale_output is None:
            unscale_output = self._output_transform_args is not None

        if scale:
            inputs = self._transform(inputs, True)

        # call the prediction
        outputs = self.model.predict(inputs, **keras_kwargs)

        # unscale if necessary
        if unscale_output:
            outputs = self._inverse_transform(outputs, is_input=False)

        return outputs


# LINN format conversion


def _eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


custom_objects = {"OutputClipLayer": OutputClipLayer, "WeightClip": WeightClip}


def convert_keras_to_linn(
    model: keras.models.Model, input_channels: int, output_channels: int, **kwargs
):
    dims: List[int] = []
    jlayers: List[dict] = []

    assert isinstance(model, (keras.Sequential, keras.Model, LinnModel)), (
        f"Unsupported model {type(model)}"
    )

    if isinstance(model, LinnModel):
        model = model.model

    colDepth = 0
    hardwareLayers = 0

    for i, layer in enumerate(model.layers):
        if isinstance(layer, (WeightClip, OutputClipLayer, keras.layers.InputLayer)):
            _eprint(f"Skipping layer {i} with type {type(layer)}")
            continue

        if isinstance(layer, keras.layers.InputLayer):
            assert i == 0, "Input layer must be first layer"
            inputs = layer.batch_shape[0]
            dims.append(inputs)
            _eprint(f"Input shape {inputs}")
        else:
            assert isinstance(layer, keras.layers.Dense), (
                f"Only Dense layers supported, got ({type(layer)})"
            )

            layerWeights = layer.get_weights()
            assert len(layerWeights) == 2, "Layer must only contains weights and biases"

            # layers are stored in column-major order
            weights = layerWeights[0].transpose()
            bias = layerWeights[1]

            assert len(weights.shape) == 2, "Only two dimensional layers supported"
            assert len(bias.shape) == 1, "Biases must be a vector"

            nRows = weights.shape[0]
            nCols = weights.shape[1]
            nBias = bias.shape[0]

            assert 0 < nBias <= 100, "Number of output rows must be 100 or fewer"
            assert nBias == nRows, "Weights output does not match bias vector size"
            if hardwareLayers == 0:
                # Handle the input to the network
                assert 1 <= nCols <= 100, "Maximum network input size is 100"
                dims.append(nCols)
            else:
                assert nCols == dims[-1], (
                    f"Input size {nCols} does not match output of previous layer {dims[-1]}"
                )
            dims.append(nRows)

            # Ensure that the total number of weights and biases is not too large
            # to fit into the memory for each neuron.
            colDepth += nCols + 3  # bias
            hardwareLayers += 1

            jlayers.append(
                {
                    "activation": layer.activation.__name__.lower(),
                    "weights": weights.tolist(),
                    "biases": bias.tolist(),
                }
            )

    assert hardwareLayers <= 5, "Only five dense layers allowed"

    assert colDepth <= 1024, (
        "Total number of weights and biases too large,"
        "sum(L[0].shape[1] for L in layers) + 3*len(layers)"
        f" = {colDepth} must be <= 1024"
    )

    _eprint(f"Network latency approx. {colDepth} cycles")

    if "output_mapping" in kwargs:
        output_map = kwargs.get("output_mapping")
        prev_final_weights = deepcopy(jlayers[-1]["weights"])
        prev_final_bias = deepcopy(jlayers[-1]["biases"])
        assert isinstance(output_map, list), "Output mapping must be a list"
        assert len(output_map) <= len(prev_final_weights), (
            f"Output mapping must have less than {len(prev_final_weights)} elements"
        )
        assert len(output_map) > 0, "Output mapping must have at least one element"

        jlayers[-1]["weights"] = [[0] * len(prev_final_weights[0])] * len(output_map)
        jlayers[-1]["biases"] = [0] * len(output_map)

        for i, output_index in enumerate(output_map):
            assert isinstance(output_index, int), "Output mapping must be a list of integers"
            assert 0 <= output_index < len(prev_final_weights), (
                "Output mapping must be in range of output size"
            )
            jlayers[-1]["weights"][i] = prev_final_weights[output_index]
            jlayers[-1]["biases"][i] = prev_final_bias[output_index]

    # Attempt to pretty print in a more readable form than json.dump; weights
    # will print as a 2D grid instead of a linear list of lists
    assert len(dims) > 1

    if len(jlayers[0]["weights"][0]) > 4:
        assert input_channels == 1, "Only one channel is supported for > 4 network inputs"
    else:
        assert input_channels == 1 or input_channels == len(jlayers[0]["weights"][0]), (
            f"Input channels must match input size {len(jlayers[0]['weights'][0])}"
        )

    if len(jlayers[-1]["weights"]) > 4:
        assert output_channels == 1, "Only one channel is supported for > 4 network outputs"
    else:
        assert output_channels == 1 or output_channels == len(jlayers[-1]["weights"]), (
            f"Output channels must match output size {len(jlayers[-1]['weights'])}"
        )

    return {
        "version": "0.1",
        "num_input_channels": input_channels,
        "num_output_channels": output_channels,
        "layers": jlayers,
    }


def get_linn(
    model: keras.models.Model, input_channels: int, output_channels: int, **kwargs
) -> dict:
    """
    Converts a 'LinnModel' into the '.linn' format required for execution on the
    Moku Neural Network Instrument. This function will also work with compatible
    Keras models if configured according to 'LinnModel' standards.

    Args:
        model (keras.models.Model): The 'LinnModel' instance or a compatible Keras model.
        input_channels: An integer of the number of instrument inputs to connect to the network.
                        Determines processing mode (serial or parallel) based on the ratio between
                        'input_channels' and the number of input neurons in the model.
        output_channels: An integer of the number of instrument outputs to connect to the network.
                        Determines processing mode (serial or parallel) based on the ratio between
                        'output_channels' and the number of output neurons in the model.
    Keyword Args (Optional):
        output_mapping (list): A list of integers that selects which output neurons
                               should be used as the final output of the network.

    Returns:
        dict: The .linn JSON document or a dict of the network parameters suitable for loading in to the Neural Network instrument.
    """
    return convert_keras_to_linn(
        model=model, input_channels=input_channels, output_channels=output_channels, **kwargs
    )


def save_linn(
    model: keras.models.Model, input_channels: int, output_channels: int, file_name: str, **kwargs
):
    """
    Converts a Keras model which is suitable for execution on the Moku
    Neural Network Instrument into the `.linn` format and saves it to a
    file.

    Args:
        model (keras.models.Model): The 'LinnModel' instance or a compatible Keras model.
        input_channels: An integer of the number of instrument inputs to connect to the network.
                        Determines processing mode (serial or parallel) based on the ratio between
                        'input_channels' and the number of input neurons in the model.
        output_channels: An integer of the number of instrument outputs to connect to the network.
                        Determines processing mode (serial or parallel) based on the ratio between
                        'output_channels' and the number of output neurons in the model.
        file_name (str): Name of output .linn file, requires .linn extension.

    Keyword Args (Optional):
        output_mapping (list): A list of integers that selects which output neurons
                               should be used as the final output of the network.
    Returns:
        None. Saves the result to a .linn file for loading in to the Neural Network instrument.
    """

    linn_data = convert_keras_to_linn(
        model=model, input_channels=input_channels, output_channels=output_channels, **kwargs
    )
    with open(file_name, "w+") as _writer:
        json.dump(linn_data, _writer)



from moku import Moku, MultiInstrumentSlottable


class Oscilloscope(MultiInstrumentSlottable, Moku):
    """
    Oscilloscope instrument object.

    The Oscilloscope instrument provides time-domain views
    of voltages. It contains a built-in Waveform Generator
    that can control the Moku analog outputs as well.

    In normal operation, the Oscilloscope shows the signal
    present on the two analog inputs but it can be set to
    loop back the signals being generated

    Read more at https://apis.liquidinstruments.com/reference/oscilloscope # noqa

    """

    INSTRUMENT_ID = 1
    OPERATION_GROUP = "oscilloscope"

    def __init__(
        self,
        ip=None,
        serial=None,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        connect_timeout=15,
        read_timeout=30,
        slot=None,
        multi_instrument=None,
        **kwargs,
    ):
        self._init_instrument(
            ip=ip,
            serial=serial,
            force_connect=force_connect,
            ignore_busy=ignore_busy,
            persist_state=persist_state,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            slot=slot,
            multi_instrument=multi_instrument,
            **kwargs,
        )

    @classmethod
    def for_slot(cls, slot, multi_instrument):
        """Configures instrument at given slot in multi instrument mode"""
        return cls(slot=slot, multi_instrument=multi_instrument)

    def save_settings(self, filename):
        """
        Save instrument settings to a file. The file name should have
        a `.mokuconf` extension to be compatible with other tools.

        :type filename: FileDescriptorOrPath
        :param filename: The path to save the `.mokuconf` file to.
        """
        self.session.get_file(f"slot{self.slot}/{self.operation_group}", "save_settings", filename)

    def load_settings(self, filename):
        """
        Load a previously saved `.mokuconf` settings file into the instrument.
        To create a `.mokuconf` file, either use `save_settings` or the desktop app.

        :type filename: FileDescriptorOrPath
        :param filename: The path to the `.mokuconf` configuration to load
        """
        with open(filename, "rb") as f:
            self.session.post_file(
                f"slot{self.slot}/{self.operation_group}", "load_settings", data=f
            )

    def set_defaults(self):
        """
        set_defaults.
        """
        operation = "set_defaults"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def sync_output_phase(self):
        """
        sync_output_phase.
        """
        operation = "sync_output_phase"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def set_frontend(self, channel, impedance, coupling, range, bandwidth=None, strict=True):
        """
        set_frontend.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type impedance: `string` ['1MOhm', '50Ohm']
        :param impedance: Input impedance

        :type coupling: `string` ['AC', 'DC']
        :param coupling: Input coupling

        :type range: `string` ['100mVpp', '400mVpp', '1Vpp', '2Vpp', '4Vpp', '10Vpp', '40Vpp', '50Vpp'] # noqa
        :param range: Input range

        :type bandwidth: `string` ['1MHz', '30MHz', '200MHz', '300MHz', '600MHz', '2GHz']
        :param bandwidth: Input bandwidth

        """
        operation = "set_frontend"
        params = dict(
            strict=strict,
            channel=channel,
            impedance=impedance,
            coupling=coupling,
            range=range,
            bandwidth=bandwidth,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_frontend(self, channel):
        """
        get_frontend.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_frontend"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_sources(self, sources, strict=True):
        """
        set_sources.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type sources: `list`
        :param sources: List of sources

        """
        operation = "set_sources"
        params = dict(
            strict=strict,
            sources=sources,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_source(self, channel, source, strict=True):
        """
        set_source.

        :type channel: `integer`
        :param channel: Target channel

        :type source: `string` ['None', 'Input1', 'Input2', 'Input3', 'Input4', 'Output1', 'Output2', 'Output3', 'Output4'] # noqa
        :param source: Set channel data source

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        """
        operation = "set_source"
        params = dict(
            channel=channel,
            source=source,
            strict=strict,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_input_attenuation(self, channel, attenuation, strict=True):
        """
        set_input_attenuation.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type attenuation: `integer`
        :param attenuation: Input attenuation(0.1 to 1000)

        """
        operation = "set_input_attenuation"
        params = dict(
            strict=strict,
            channel=channel,
            attenuation=attenuation,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_sources(self):
        """
        get_sources.
        """
        operation = "get_sources"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_interpolation(self, interpolation="SinX", strict=True):
        """
        set_interpolation.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type interpolation: `string` ['Linear', 'SinX', 'Gaussian']
        :param interpolation: Interpolation

        """
        operation = "set_interpolation"
        params = dict(strict=strict, interpolation=interpolation)
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_interpolation(self):
        """
        get_interpolation.
        """
        operation = "get_interpolation"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def disable_input(self, channel, strict=True):
        """
        disable_input.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "disable_input"
        params = dict(
            strict=strict,
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_trigger(
        self,
        type="Edge",
        level=0,
        level_low=0,
        level_high=0.1,
        mode="Auto",
        edge="Rising",
        polarity="Positive",
        width=0.0001,
        width_condition="LessThan",
        nth_event=1,
        holdoff=0,
        hysteresis=1e-3,
        auto_sensitivity=True,
        noise_reject=False,
        hf_reject=False,
        source="Input1",
        strict=True,
    ):
        """
        set_trigger.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type type: `string` ['Edge', 'Pulse', 'Runt']
        :param type: Trigger type

        :type level: `number` [-5V, 5V]  (defaults to 0)
        :param level: Trigger level

        :type levelLow: `number` [-5V, 5V]  (defaults to 0)
        :param level: Low trigger level (Runt mode only)

        :type levelHigh: `number` [-5V, 5V]  (defaults to 0.1)
        :param level: High trigger level (Runt mode only)

        :type mode: `string` ['Auto', 'Normal']
        :param mode: Trigger mode

        :type edge: `string` ['Rising', 'Falling', 'Both']
        :param edge: Which edge to trigger on (Edge and Pulse mode). In Pulse Width modes this specifies whether the pulse is positive (rising) or negative (falling), with the 'both' option being invalid # noqa

        :type polarity: `string` ['Positive', 'Negative']
        :param polarity: Trigger pulse polarity (Runt and Pulse mode)

        :type width: `number` [26e-3Sec, 10Sec]  (defaults to 0.0001)
        :param width: Width of the trigger pulse (Pulse mode only)

        :type width_condition: `string` ['GreaterThan', 'LessThan']
        :param width_condition: Trigger pulse width condition (pulse mode only)

        :type nth_event: `integer` [0, 65535]  (defaults to 1)
        :param nth_event: The number of trigger events to wait for before triggering # noqa

        :type holdoff: `number` [1e-9Sec, 10Sec]  (defaults to 0)
        :param holdoff: The duration to hold-off Oscilloscope trigger post trigger event # noqa

        :type hysteresis: `number` (defaults to 0.001)
        :param hysteresis: Absolute hysteresis around trigger

        :type auto_sensitivity: `boolean`
        :param auto_sensitivity: Configure auto or manual hysteresis for noise rejection.

        :type noise_reject: `boolean`
        :param noise_reject: Configure the Oscilloscope with a small amount of hysteresis to prevent repeated triggering due to noise # noqa

        :type hf_reject: `boolean`
        :param hf_reject: Configure the trigger signal to pass through a low pass filter to smooths out the noise # noqa

        :type source: `string` ['ChannelA', 'ChannelB', 'ChannelC', 'ChannelD', 'Input1', 'Input2', 'Input3', 'Input4', 'Output1', 'Output2', 'Output3', 'Output4', 'External'] # noqa
        :param source: Trigger Source

        """
        operation = "set_trigger"
        params = dict(
            strict=strict,
            type=type,
            level=level,
            level_low=level_low,
            level_high=level_high,
            mode=mode,
            edge=edge,
            polarity=polarity,
            width=width,
            width_condition=width_condition,
            nth_event=nth_event,
            holdoff=holdoff,
            hysteresis=hysteresis,
            auto_sensitivity=auto_sensitivity,
            noise_reject=noise_reject,
            hf_reject=hf_reject,
            source=source,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def osc_measurement(self, t1, t2, trigger_source, edge, level, strict=True):
        """
        osc_measurement.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type t1: `number`
        :param t1: Time from the trigger point to the left of screen

        :type t2: `number`
        :param t2: Time from the trigger point to the right of screen. (Must be a positive number, i.e. after the trigger event) # noqa

        :type trigger_source: `string` ['ChannelA', 'ChannelB', 'ChannelC', 'ChannelD', 'Input1', 'Input2', 'Input3', 'Input4', 'Output1', 'Output2', 'Output3', 'Output4', 'External'] # noqa
        :param trigger_source: Trigger source

        :type edge: `string` ['Rising', 'Falling', 'Both']
        :param edge: Which edge to trigger on. Only edge trigger is used with this function, pulse trigger can be enabled using set_trigger() # noqa

        :type level: `number` [-5V, 5V]
        :param level: Trigger level

        """
        operation = "osc_measurement"
        params = dict(
            strict=strict,
            t1=t1,
            t2=t2,
            trigger_source=trigger_source,
            edge=edge,
            level=level,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def summary(self):
        """
        summary.
        """
        operation = "summary"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_timebase(self, t1, t2, max_length=1024, frame_length=None, strict=True):
        """
        set_timebase.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type t1: `number`
        :param t1: Time from the trigger point to the left of screen. This may be negative (trigger on-screen) or positive (trigger off the left of screen). # noqa

        :type t2: `number`
        :param t2: Time from the trigger point to the right of screen. (Must be a positive number, i.e. after the trigger event) # noqa

        :type max_length: `number`
        :param max_length: Maximum length of the frame to retrieve through get_data(). The API will attempt to return a frame with as many points as possible, up to this limit. # noqa

        :type frame_length: `number`
        :param frame_length: Deprecated, use `max_length` instead which better describes the real behaviour of this parameter.

        """
        operation = "set_timebase"
        params = dict(
            strict=strict, t1=t1, t2=t2, max_length=max_length, frame_length=frame_length
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_hysteresis(self, hysteresis_mode, value=0, strict=True):
        """
        set_hysteresis.

        .. deprecated:: 3.1.1
        Use `hysteresis` parameter of `set_trigger` instead.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type hysteresis_mode: `string` ['Absolute', 'Relative']
        :param hysteresis_mode: Trigger sensitivity hysteresis mode

        :type value: `number`
        :param value: Hysteresis around trigger

        """
        operation = "set_hysteresis"
        params = dict(
            strict=strict,
            hysteresis_mode=hysteresis_mode,
            value=value,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def enable_rollmode(self, roll=True, strict=True):
        """
        enable_rollmode.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type roll: `boolean`
        :param roll: Enable roll

        """
        operation = "enable_rollmode"
        params = dict(
            strict=strict,
            roll=roll,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_data(self, timeout=60, wait_reacquire=False, wait_complete=False, measurements=False):
        """
        get_data.

        Returns a dictionary containing the data frame and the timebase. The length of the data frame
        defaults to 1024 points but can be changed using the set_timebase() function. Changes to the
        requested frame length are given as the maximum number of points that the user will accept,
        the actual length may be slightly less than this. The actual length may also change depending
        on the specific timebase, so it's recommended to write code that is able to handle variable
        frame lengths.

        :type timeout: `number` Seconds (defaults to 60)
        :param timeout: Wait for n seconds to receive a data frame

        :type wait_reacquire: `boolean`
        :param wait_reacquire: Wait until new dataframe is reacquired

        :type wait_complete: `boolean`
        :param wait_complete: Wait until entire frame is available

        :type measurements: `boolean`
        :param measurements: When True, includes available measurements for each channel


        .. important::
            Default timeout for reading the data is 10 seconds. It
            can be increased by setting the read_timeout property of
            session object.

            Example: ``i.session.read_timeout=100`` (in seconds)

        """
        operation = "get_data"
        params = dict(
            timeout=timeout,
            wait_reacquire=wait_reacquire,
            wait_complete=wait_complete,
            measurements=measurements,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def save_high_res_buffer(self, comments="", timeout=60):
        """
        save_high_res_buffer.

        :type comments: `string`
        :param comments: Optional comments to be included

        :type timeout: `int`
        :param timeout: Wait for n seconds before trigger event happens to save the buffer # noqa

        """
        operation = "save_high_res_buffer"
        params = dict(comments=comments, timeout=timeout)
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_acquisition_mode(self, mode="Normal", strict=True):
        """
        set_acquisition_mode.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type mode: `string` ['Normal', 'Precision', 'DeepMemory', 'PeakDetect'] # noqa
        :param mode: Acquisition Mode

        """
        operation = "set_acquisition_mode"
        params = dict(strict=strict, mode=mode)
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_samplerate(self):
        """
        get_samplerate.
        """
        operation = "get_samplerate"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def get_acquisition_mode(self):
        """
        get_acquisition_mode.
        """
        operation = "get_acquisition_mode"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def get_timebase(self):
        """
        get_timebase.
        """
        operation = "get_timebase"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def generate_waveform(
        self,
        channel,
        type,
        amplitude=1,
        frequency=10000,
        offset=0,
        phase=0,
        duty=None,
        symmetry=None,
        dc_level=None,
        edge_time=None,
        pulse_width=None,
        strict=True,
    ):
        """
        generate_waveform.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type type: `string` ['Off', 'Sine', 'Square', 'Ramp', 'Pulse', 'Noise', 'DC'] # noqa
        :param type: Waveform type

        :type amplitude: `number` [4e-3V, 10V]  (defaults to 1)
        :param amplitude: Waveform peak-to-peak amplitude

        :type frequency: `number` [1e-3Hz, 20e6Hz]  (defaults to 10000)
        :param frequency: Waveform frequency

        :type offset: `number` [-5V, 5V]  (defaults to 0)
        :param offset: DC offset applied to the waveform

        :type phase: `number` [0Deg, 360Deg]  (defaults to 0)
        :param phase: Waveform phase offset

        :type duty: `number` [0.0%, 100.0%]
        :param duty: Duty cycle as percentage (Only for Square wave)

        :type symmetry: `number` [0.0%, 100.0%]
        :param symmetry: Fraction of the cycle rising

        :type dc_level: `number`
        :param dc_level: DC Level. (Only for DC waveform)

        :type edge_time: `number` [16e-9, pulse width]
        :param edge_time: Edge-time of the waveform (Only for Pulse wave)

        :type pulse_width: `number`
        :param pulse_width: Pulse width of the waveform (Only for Pulse wave)

        """
        operation = "generate_waveform"
        params = dict(
            strict=strict,
            channel=channel,
            type=type,
            amplitude=amplitude,
            frequency=frequency,
            offset=offset,
            phase=phase,
            duty=duty,
            symmetry=symmetry,
            dc_level=dc_level,
            edge_time=edge_time,
            pulse_width=pulse_width,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_output_load(self, channel, load, strict=True):
        """
        .. deprecated:: 3.1.1
        Use `set_output_termination` instead.

        set_output_load.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type load: `string` ['1MOhm', '50Ohm']
        :param load: Output load

        """
        operation = "set_output_load"
        params = dict(
            strict=strict,
            channel=channel,
            load=load,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output_load(self, channel):
        """
        .. deprecated:: 3.1.1
        Use `get_output_termination` instead.

        get_output_load.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output_load"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_output_termination(self, channel, termination, strict=True):
        """
        set_output_termination.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type termination: `string` ['HiZ', '50Ohm']
        :param termination: Output termination

        """
        operation = "set_output_termination"
        params = dict(
            strict=strict,
            channel=channel,
            termination=termination,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output_termination(self, channel):
        """
        get_output_termination.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output_termination"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)



import json
import subprocess
import tarfile
import warnings
from os import environ
from pathlib import Path
from shutil import which

from moku.exceptions import (
    IncompatibleMokuException,
    IncompatiblePackageException,
    MokuException,
    MokuNotFound,
    NoInstrumentBitstream,
)
from moku.session import RequestSession
from moku.version import COMPAT_MOKUOS, SUPPORTED_PROXY_VERSION
from requests.exceptions import ConnectionError

from .utilities import check_mokucli_version, get_bitstream_path, get_config_dir
from .logging import get_logger

# Set up logger for this module
logger = get_logger(__name__.split(".")[-1])

# Look for mokucli in the path or in the MOKU_CLI_PATH environment variable
if MOKU_CLI_PATH := environ.get("MOKU_CLI_PATH", which("mokucli")):
    MOKU_CLI_PATH = str(Path(MOKU_CLI_PATH).expanduser())
else:
    warnings.warn(
        "Can't find mokucli, which is required for correct operation. Please install from https://liquidinstruments.com/software/utilities/ or set the MOKU_CLI_PATH environment variable to the path of the mokucli executable. Using default path."
    )

# The MOKU_DATA_PATH environment variable is used to override the default data path,
# which is picked up from the mokucli or by using the same logic as mokucli
# which we reproduce in the utilities.py module.
if MOKU_DATA_PATH := environ.get("MOKU_DATA_PATH"):
    MOKU_DATA_PATH = Path(MOKU_DATA_PATH).expanduser()
else:
    if MOKU_CLI_PATH:
        _data_path = (
            subprocess.check_output([MOKU_CLI_PATH, "config", "which"]).decode("utf-8").strip()
        )
        MOKU_DATA_PATH = Path(_data_path).parent.joinpath("data")
    else:
        MOKU_DATA_PATH = get_config_dir().joinpath("data")


class MultiInstrumentSlottable:
    """Mixin to handle common instrument initialization pattern for multi-instrument capable devices.

    Must mix in to a class that also extends Moku in order to get bitstream upload implementation.
    """

    # Subclasses must define these
    INSTRUMENT_ID = None
    OPERATION_GROUP = None

    def _init_instrument(
        self,
        ip=None,
        serial=None,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        connect_timeout=15,
        read_timeout=30,
        slot=None,
        multi_instrument=None,
        bs_path=None,
        **kwargs,
    ):
        """Common initialization logic for all instruments.

        Args:
            bs_path: Optional path to custom bitstream file. If provided, this will be
                    used instead of the standard instrument bitstream.
        """

        if self.INSTRUMENT_ID is None or self.OPERATION_GROUP is None:
            raise NotImplementedError("Subclass must define INSTRUMENT_ID and OPERATION_GROUP")

        # Get instrument-specific logger
        self._logger = get_logger(f"instruments.{self.__class__.__name__}")
        self._logger.debug(f"Initializing {self.__class__.__name__} (ID={self.INSTRUMENT_ID})")

        self.id = self.INSTRUMENT_ID
        self.operation_group = self.OPERATION_GROUP

        if multi_instrument is None:
            # Standalone mode
            self.slot = 1
            self._logger.debug("Initializing in standalone mode (slot=1)")
            if not any([ip, serial]):
                raise MokuException("IP (or) Serial is required")
            if serial:
                from moku.utilities import find_moku_by_serial

                self._logger.debug(f"Finding Moku by serial: {serial}")
                ip = find_moku_by_serial(serial)

            # Call Moku.__init__
            super().__init__(
                ip=ip,
                force_connect=force_connect,
                ignore_busy=ignore_busy,
                persist_state=persist_state,
                connect_timeout=connect_timeout,
                read_timeout=read_timeout,
                **kwargs,
            )
            self.upload_bitstream("01-000")
            self.upload_bitstream(f"01-{self.id:03}-00", bs_path=bs_path)
        else:
            # Multi-instrument mode
            self.platform_id = multi_instrument.platform_id
            self.slot = slot
            self.session = multi_instrument.session
            self.mokuOS_version = multi_instrument.mokuOS_version
            self.hardware = multi_instrument.hardware
            self.bitstreams = multi_instrument.bitstreams
            self.manage_bitstreams = multi_instrument.manage_bitstreams
            self._logger.info(
                f"Uploading bitstream for {self.__class__.__name__} in slot {self.slot}"
            )
            self.upload_bitstream(
                f"{self.platform_id:02}-{self.id:03}-{self.slot - 1:02}", bs_path=bs_path
            )
            self._logger.debug(f"Deploying instrument to slot {self.slot}")
            self.session.get(f"slot{self.slot}", self.operation_group)
            self._logger.info(
                f"{self.__class__.__name__} successfully deployed to slot {self.slot}"
            )


class Moku:
    """
    Moku base class. This class does all the heavy lifting required to
    deploy and control instruments.
    """

    def __init__(
        self,
        ip,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        connect_timeout=15,
        read_timeout=30,
        **kwargs,
    ) -> None:
        self._am_owner = False
        logger.info(f"Initializing Moku connection to {ip}")
        logger.debug(
            f"Connection parameters: force_connect={force_connect}, ignore_busy={ignore_busy}, persist_state={persist_state}"
        )

        try:
            self.session = RequestSession(ip, connect_timeout, read_timeout, **kwargs)
            logger.debug("Session created, claiming ownership")
            force_deploy = kwargs.get("force_deploy", False)
            self.claim_ownership(force_connect, force_deploy, ignore_busy, persist_state)

            props = self.describe()
            logger.debug(
                f"Device properties: hardware={props.get('hardware')}, mokuOS={props.get('mokuOS')}"
            )

            if kwargs.get("no_check_version", False):
                pass
            else:
                # TODO: Split the CLI version check from the MokuOS version check?
                if MOKU_CLI_PATH:
                    # If mokucli is found, check it's compatible. If it's not found
                    # then we're likely to fail later like uploading bitstreams.
                    check_mokucli_version(MOKU_CLI_PATH)
                if "proxy_version" not in props:
                    raise IncompatibleMokuException(
                        f"Incompatible MokuOS version, this version of "
                        f"package supports MokuOS {COMPAT_MOKUOS}."
                    )
                if int(props["proxy_version"]) > SUPPORTED_PROXY_VERSION:
                    raise IncompatiblePackageException(
                        "Incompatible Moku package, please update the package using pip"
                    )
                elif int(props["proxy_version"]) < SUPPORTED_PROXY_VERSION:
                    raise IncompatibleMokuException(
                        "You are using an old version of MokuOS. "
                        "Please update to the latest version using the Moku Desktop app or mokucli."
                    )
        except (IncompatibleMokuException, IncompatiblePackageException):
            self.relinquish_ownership()
            raise
        except ConnectionError as e:
            self.relinquish_ownership()
            logger.error(f"Connection failed to {ip}: {e}")
            raise MokuNotFound("Could not connect to Moku")
        except Exception as e:
            self.relinquish_ownership()
            logger.error(f"Unexpected error during initialization: {e}")
            raise MokuException(f"An unexpected error occurred during initialization: {e}")

        self.mokuOS_version: str = props["mokuOS"]
        self.hardware: str = props["hardware"].replace(":", "").lower()
        self.bitstreams = props["bitstreams"]
        self.manage_bitstreams = kwargs.get("manage_bitstreams", True)
        logger.info(
            f"Successfully connected to {self.hardware} running MokuOS {self.mokuOS_version}"
        )

    def _upload_bitstream_if_required(self, bs_name, rmt_chksum, bs_path=None) -> None:
        logger.debug(
            f"Checking bitstream {bs_name} (remote checksum: {rmt_chksum[:8] if rmt_chksum else 'None'}...)"
        )
        try:
            bs_path = Path(bs_path or get_bitstream_path(self.mokuOS_version, self.hardware))
            bs_file_name = bs_path / bs_name
            if not bs_file_name.exists():
                logger.error(f"Bitstream file not found: {bs_file_name}")
                raise MokuException(f"Cannot find {bs_file_name}")

            with tarfile.open(bs_file_name, mode="r") as _bar:
                if "MANIFEST" not in _bar.getnames():
                    raise NoInstrumentBitstream(
                        f"MANIFEST file is missing in the bitstream {bs_file_name}."
                    )
                bs_man_file = _bar.extractfile("MANIFEST")
                if not bs_man_file:
                    raise NoInstrumentBitstream(
                        f"Failed to extract MANIFEST file from the bitstream {bs_file_name}."
                    )
                bs_manifest = json.loads(bs_man_file.read())
                local_chksum = bs_manifest["items"][0]["sha256"]
                if not (rmt_chksum and local_chksum == rmt_chksum):
                    logger.info(f"Uploading bitstream {bs_name} (checksum mismatch or missing)")
                    self.upload("bitstreams", bs_name, open(bs_file_name, "rb").read())
                else:
                    logger.debug(f"Bitstream {bs_name} already up to date")

        except (NoInstrumentBitstream, MokuException):
            self.relinquish_ownership()
            raise NoInstrumentBitstream(
                f"Instrument files not available, please run `mokucli instrument download {self.mokuOS_version}` to download latest instrument data"
            )

        except Exception as e:
            self.relinquish_ownership()
            raise MokuException(f"An unexpected error occurred while uploading bitstream: {e}")

    def upload_bitstream(self, name, bs_path=None):
        if self.manage_bitstreams:
            name = f"{name}.bar"
            exists = [b[1] for b in self.bitstreams.items() if b[0] == name]
            rmt_chksum = exists[0] if exists else None
            self._upload_bitstream_if_required(name, rmt_chksum, bs_path)

    def set_connect_timeout(self, value):
        "Sets requests session connect timeout"
        if not isinstance(value, tuple([int, float])):
            raise ValueError("set_connect_timeout value should be either integer or float")
        self.session.connect_timeout = value

    def set_read_timeout(self, value):
        "Sets requests session read timeout"
        if not isinstance(value, tuple([int, float])):
            raise ValueError("read_timeout value should be either integer or float")
        self.session.read_timeout = value

    def platform(self, platform_id):
        "Configures platform for the given ID"
        operation = f"platform/{platform_id}"
        self.upload_bitstream(f"{platform_id:02}-000")
        for i in range(0, platform_id):
            self.upload_bitstream(f"{platform_id:02}-000-{i:02}")
        return self.session.get("moku", operation)

    def claim_ownership(
        self, force_connect=True, force_deploy=False, ignore_busy=False, persist_state=False
    ):
        """
        Claim the ownership of Moku.

        :type force_connect: `boolean`
        :param force_connect: Force connection to Moku disregarding any existing connections

        :type force_deploy: `boolean`
        :param force_deploy: Reload platform bitstream when starting an instrument

        :type ignore_busy: `boolean`
        :param ignore_busy: Ignore the state of instrument including any in progress data logging sessions and proceed with the deployment # noqa

        :type persist_state: `boolean`
        :param persist_state: When true, tries to retain the previous state of the instrument(if available) # noqa

        """
        operation = "claim_ownership"
        params = dict(
            force_connect=force_connect,
            force_deploy=force_deploy,
            ignore_busy=ignore_busy,
            persist_state=persist_state,
        )
        logger.debug(f"Claiming ownership with params: {params}")
        ret = self.session.post("moku", operation, params)
        self._am_owner = True
        logger.info("Successfully claimed ownership of Moku")
        return ret

    def relinquish_ownership(self):
        """
        Relinquish the ownership of Moku.
        """
        operation = "relinquish_ownership"
        if hasattr(self, "_am_owner") and self._am_owner:
            logger.debug("Relinquishing ownership")
            ret = self.session.post("moku", operation)
            self._am_owner = False
            logger.info("Successfully relinquished ownership")
        else:
            logger.debug("Not owner, skipping relinquish")
            ret = None
        return ret

    def __enter__(self):
        """
        Enter the runtime context for the Moku object.

        Returns the Moku instance for use in the with statement.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the runtime context for the Moku object.

        Automatically relinquishes ownership when exiting the context.
        """
        if exc_type:
            logger.debug(f"Exiting context with exception: {exc_type.__name__}: {exc_value}")
        else:
            logger.debug("Exiting context normally")
        self.relinquish_ownership()
        return False

    def name(self):
        """
        name.
        """
        operation = "name"
        return self.session.get("moku", operation)

    def serial_number(self):
        """
        serial_number.
        """
        operation = "serial_number"
        return self.session.get("moku", operation)

    def summary(self):
        """
        summary.
        """
        operation = "summary"
        return self.session.get("moku", operation)

    def describe(self):
        """
        describe.
        """
        operation = "describe"
        return self.session.get("moku", operation)

    def calibration_date(self):
        """
        calibration_date.
        """
        operation = "calibration_date"
        return self.session.get("moku", operation)

    def firmware_version(self):
        """
        .. deprecated:: 3.4.0
        MokuOS version.
        """
        print("Warning: firmware_version() is deprecated, use mokuos_version() instead")
        return self.mokuos_version()

    def mokuos_version(self):
        """
        MokuOS version.
        """
        operation = "mokuos_version"
        return self.session.get("moku", operation)

    def get_power_supplies(self):
        """
        get_power_supplies.
        """
        operation = "get_power_supplies"
        return self.session.get("moku", operation)

    def get_power_supply(self, id):
        """
        get_power_supply.

        :type id: `integer`
        :param id: ID of the power supply

        """
        operation = "get_power_supply"
        params = dict(
            id=id,
        )
        return self.session.post("moku", operation, params)

    def set_power_supply(self, id, enable=True, voltage=3, current=0.1):
        """
        set_power_supply.

        :type id: `integer`
        :param id: ID of the power supply to configure

        :type enable: `boolean`
        :param enable: Enable/Disable power supply

        :type voltage: `number`
        :param voltage: Voltage set point

        :type current: `number`
        :param current: Current set point

        """
        operation = "set_power_supply"
        params = dict(
            id=id,
            enable=enable,
            voltage=voltage,
            current=current,
        )
        return self.session.post("moku", operation, params)

    def get_external_clock(self):
        """
        get_external_clock.
        """
        operation = "get_external_clock"
        return self.session.get("moku", operation)

    def set_external_clock(self, enable=True):
        """
        set_external_clock.

        :type enable: `boolean`
        :param enable: Switch between external and internal reference clocks

        """
        operation = "set_external_clock"
        params = dict(
            enable=enable,
        )
        return self.session.post("moku", operation, params)

    def get_blended_clock(self):
        """
        get_blended_clock.
        """
        operation = "get_blended_clock"
        return self.session.get("moku", operation)

    def set_blended_clock(
        self,
        freq_ref_enable=None,
        freq_ref_frequency=None,
        sync_ref_enable=None,
        sync_ref_source=None,
        strict=True,
    ):
        """
        set_blended_clock.

        :type freq_ref_enable: `boolean`
        :param freq_ref_enable: Enable external frequency reference

        :type freq_ref_frequency: `string` ['10MHz', '100MHz']
        :param freq_ref_frequency: Frequency of external frequency reference

        :type sync_ref_enable: `boolean`
        :param sync_ref_enable: Enable synchronization reference

        :type sync_ref_source: `string` [ 'GNSS', 'Ext']
        :param sync_ref_source: Source of synchronization reference

        """
        operation = "set_blended_clock"
        params = dict(
            freq_ref_enable=freq_ref_enable,
            freq_ref_frequency=freq_ref_frequency,
            sync_ref_enable=sync_ref_enable,
            sync_ref_source=sync_ref_source,
        )
        return self.session.post("moku", operation, params)

    def upload(self, target, file_name, data):
        """
        Upload files to bitstreams, ssd, logs, persist.

        :type target: `string`, (bitstreams, ssd, logs, persist, media)
        :param target: Destination where the file should be uploaded to.

        :type file_name: `string`
        :param file_name: Name of the file to be uploaded

        :type data: `bytes`
        :param data: File content

        """
        operation = f"upload/{file_name}"
        return self.session.post_file(target, operation, data)

    def delete(self, target, file_name):
        """
        Delete files from bitstreams, ssd, logs, persist.

        :type target: `string`, (bitstreams, ssd, logs, persist, media)
        :param target: Destination where the file should be uploaded to.

        :type file_name: `string`
        :param file_name: Name of the file to be deleted

        """
        operation = f"delete/{file_name}"
        return self.session.delete_file(target, operation)

    def list(self, target):
        """
        List files at bitstreams, ssd, logs, persist.

        :type target: `string`, (bitstreams, ssd, logs, persist, media)
        :param target: Target directory to list files for

        """
        operation = "list"
        return self.session.get(target, operation)

    def download(self, target, file_name, local_path):
        """
        Download files from bitstreams, ssd, logs, persist.

        :type target: `string`, (bitstreams, ssd, logs, persist, media)
        :param target: Destination where the file should be downloaded from.

        :type file_name: `string`
        :param file_name: Name of the file to be downloaded

        :type local_path: `string`
        :param local_path: Local path to download the file

        """
        operation = f"download/{file_name}"
        return self.session.get_file(target, operation, local_path)

    def modify_hardware(self, data=None):
        """
        CAUTION: Never use to update the state of the Moku
        Raw access to Moku hardware state
        """
        if data is None:
            data = {}
        return self.session.post("moku", "modify_hardware", data)

    def modify_calibration(self, data=None):
        """
        Query or update the calibration coefficients
        """
        if data is None:
            data = {}
        return self.session.post("moku", "modify_calibration", data)

    def set_configuration(self, data=None):
        """
        Update the Moku device/network configuration.
        """
        if data is None:
            data = {}
        return self.session.post("moku", "modify_calibration", data)

    def get_configuration(self):
        """
        Retreive the Moku device/network configuration.
        """
        return self.session.get("moku", "get_configuration")

    def shutdown(self):
        """
        Shutdown the Moku device.
        """
        return self.session.post_to_v2("hwstate", {"power": "shutdown"})["power"]

    def reboot(self):
        """
        Reboot the Moku device.
        """
        return self.session.post_to_v2("hwstate", {"power": "reboot"})["power"]



import json
import os
from pathlib import Path
import platform
from subprocess import PIPE, Popen

from packaging.specifiers import SpecifierSet

from .exceptions import InvalidParameterRange, MokuException, MokuNotFound, NoInstrumentBitstream
from .finder import Finder
from .version import COMPAT_MOKUCLI

from .logging import get_logger

logger = get_logger(__name__.split(".")[-1])


def find_moku_by_serial(serial):
    result = Finder().find_all(timeout=10, filter=lambda x: x.serial == serial)
    if len(result) > 0:
        return result[0].ipv4_addr
    raise MokuNotFound()


def check_mokucli_version(cli_path):
    req_ver = SpecifierSet(COMPAT_MOKUCLI)
    out, _ = Popen([cli_path, "--version"], stdout=PIPE, stderr=PIPE).communicate()
    if not out:
        raise MokuException(
            "Cannot find mokucli. \n"
            "Please download latest version of the CLI from https://www.liquidinstruments.com/software/utilities/. \n"  # noqa
            "If you have already installed, please set the MOKU_CLI_PATH environment variable to absolute path of mokucli."
        )

    ver_str = out.decode("utf8").rstrip()
    if not req_ver.contains(ver_str):
        raise InvalidParameterRange(
            f"mokucli version {ver_str} is not compatible with this version of the API. \n"
            f"The minimum required version is {req_ver}. \n"
            "Please download latest version of the CLI from https://www.liquidinstruments.com/software/utilities/. \n"  # noqa
            "If you have already installed, please set the MOKU_CLI_PATH environment variable to absolute path of mokucli."
        )


def get_config_dir() -> Path:
    """Get the platform-specific configuration directory.

    This path resolution should exactly match the logic in mokucli.

    Corrolary: If you change this, change it in mokucli as well.
    """
    if platform.system() == "Windows":
        # If APPDATA is not set, use the user's home directory
        return Path(os.environ.get("APPDATA", Path.home())) / "Moku"
    elif platform.system() == "Darwin":  # macOS
        return Path.home() / "Library" / "Application Support" / "Moku"
    else:  # Linux and others
        return Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")) / "moku"


def get_version_info(mokuOS_version):
    version_file = get_config_dir() / "data" / "versions" / f"{mokuOS_version}.json"
    logger.debug(f"Looking for version info in {version_file}")
    if not version_file.exists():
        raise NoInstrumentBitstream()
    with open(version_file, "r") as f:
        return json.load(f)


def get_bitstream_path(mokuOS_version, hardware):
    hw_dir = {"mokupro": "mokupro", "mokugo": "mokugo", "mokulab": "moku20", "mokudelta": "mokuaf"}
    version_info = get_version_info(mokuOS_version)
    bitstream_path = (
        get_config_dir() / "data" / "instruments" / version_info["instruments"] / hw_dir[hardware]
    )
    logger.debug(f"Bitstream path: {bitstream_path}")
    if not bitstream_path.exists():
        raise NoInstrumentBitstream()
    return bitstream_path



from moku import Moku, MultiInstrumentSlottable


class FrequencyResponseAnalyzer(MultiInstrumentSlottable, Moku):
    """
    FrequencyResponseAnalyzer instrument object.

    This instrument measures the transfer function of a
    system by generating a swept sine wave and measuring
    the system response on the input

    Read more at https://apis.liquidinstruments.com/reference/fra

    """

    INSTRUMENT_ID = 9
    OPERATION_GROUP = "fra"

    def __init__(
        self,
        ip=None,
        serial=None,
        force_connect=False,
        ignore_busy=False,
        persist_state=False,
        connect_timeout=15,
        read_timeout=30,
        slot=None,
        multi_instrument=None,
        **kwargs,
    ):
        self._init_instrument(
            ip=ip,
            serial=serial,
            force_connect=force_connect,
            ignore_busy=ignore_busy,
            persist_state=persist_state,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            slot=slot,
            multi_instrument=multi_instrument,
            **kwargs,
        )

    @classmethod
    def for_slot(cls, slot, multi_instrument):
        """Configures instrument at given slot in multi instrument mode"""
        return cls(slot=slot, multi_instrument=multi_instrument)

    def save_settings(self, filename):
        """
        Save instrument settings to a file. The file name should have
        a `.mokuconf` extension to be compatible with other tools.

        :type filename: FileDescriptorOrPath
        :param filename: The path to save the `.mokuconf` file to.
        """
        self.session.get_file(f"slot{self.slot}/{self.operation_group}", "save_settings", filename)

    def load_settings(self, filename):
        """
        Load a previously saved `.mokuconf` settings file into the instrument.
        To create a `.mokuconf` file, either use `save_settings` or the desktop app.

        :type filename: FileDescriptorOrPath
        :param filename: The path to the `.mokuconf` configuration to load
        """
        with open(filename, "rb") as f:
            self.session.post_file(
                f"slot{self.slot}/{self.operation_group}", "load_settings", data=f
            )

    def summary(self):
        """
        summary.
        """
        operation = "summary"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_defaults(self):
        """
        set_defaults.
        """
        operation = "set_defaults"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def set_frontend(self, channel, impedance, coupling, range, bandwidth=None, strict=True):
        """
        set_frontend.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type impedance: `string` ['1MOhm', '50Ohm']
        :param impedance: Input impedance

        :type coupling: `string` ['AC', 'DC']
        :param coupling: Input coupling

        :type range: `string` ['100mVpp', '400mVpp', '1Vpp', '2Vpp', '4Vpp', '10Vpp', '40Vpp', '50Vpp'] # noqa
        :param range: Input range

        :type bandwidth: `string` ['1MHz', '30MHz', '200MHz', '300MHz', '600MHz', '2GHz']
        :param bandwidth: Input bandwidth

        """
        operation = "set_frontend"
        params = dict(
            strict=strict,
            channel=channel,
            impedance=impedance,
            coupling=coupling,
            range=range,
            bandwidth=bandwidth,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_frontend(self, channel):
        """
        get_frontend.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_frontend"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def fra_measurement(
        self,
        channel,
        mode="InOut",
        start_frequency=0,
        stop_frequency=0,
        averaging_duration=0,
        averaging_cycles=0,
        output_amplitude=0,
        strict=True,
    ):
        """
        fra_measurement.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type mode: `string` ['In', 'InOut', 'InIn1']
        :param mode: FRA Measurement mode

        :type start_frequency: `number` [1e-3Hz, 20e6Hz]  (defaults to 0)
        :param start_frequency: Sweep start frequency

        :type stop_frequency: `number` [1e-3Hz, 20e6Hz]  (defaults to 0)
        :param stop_frequency: Sweep end frequency

        :type averaging_duration: `number` [1e-6Sec, 10Sec]  (defaults to 0)
        :param averaging_duration: Minimum averaging time per sweep point.

        :type averaging_cycles: `integer` [1, 1048576]  (defaults to 0)
        :param averaging_cycles: Minimum averaging cycles per sweep point.

        :type output_amplitude: `number` [2e-3Vpp, 10Vpp]  (defaults to 0)
        :param output_amplitude: Output amplitude

        """
        operation = "fra_measurement"
        params = dict(
            strict=strict,
            channel=channel,
            mode=mode,
            start_frequency=start_frequency,
            stop_frequency=stop_frequency,
            averaging_duration=averaging_duration,
            averaging_cycles=averaging_cycles,
            output_amplitude=output_amplitude,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def measurement_mode(self, mode="InOut", strict=True):
        """
        measurement_mode.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type mode: `string` ['In', 'InOut', 'InIn1']
        :param mode: FRA Measurement mode

        """
        operation = "measurement_mode"
        params = dict(
            strict=strict,
            mode=mode,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_sweep(
        self,
        start_frequency=0,
        stop_frequency=0,
        num_points=512,
        averaging_time=0,
        averaging_cycles=0,
        settling_time=0,
        settling_cycles=0,
        dynamic_amplitude=False,
        linear_scale=False,
        strict=True,
    ):
        """
        set_sweep.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type start_frequency: `number` [1e-3Hz, 20e6Hz]  (defaults to 0)
        :param start_frequency: Sweep start frequency

        :type stop_frequency: `number` [1e-3Hz, 20e6Hz]  (defaults to 0)
        :param stop_frequency: Sweep stop frequency

        :type num_points: `integer`
        :param num_points: Number of points in the sweep (rounded to nearest power of 2) # noqa

        :type averaging_time: `number` [1e-6Sec, 10Sec]  (defaults to 0)
        :param averaging_time: Minimum averaging time per sweep point.

        :type averaging_cycles: `integer` [1, 1048576]  (defaults to 0)
        :param averaging_cycles: Minimum averaging cycles per sweep point.

        :type settling_time: `number` [1e-6Sec, 10Sec]  (defaults to 0)
        :param settling_time: Minimum settling time per sweep point.

        :type settling_cycles: `integer` [1, 1048576]  (defaults to 0)
        :param settling_cycles: Minimum settling cycles per sweep point.

        :type linear_scale: `boolean`
        :param linear_scale: Enables linear scale. If set to false scale is set to logarithmic. Defaults to false # noqa

        """
        operation = "set_sweep"
        params = dict(
            strict=strict,
            start_frequency=start_frequency,
            stop_frequency=stop_frequency,
            num_points=num_points,
            averaging_time=averaging_time,
            averaging_cycles=averaging_cycles,
            settling_time=settling_time,
            settling_cycles=settling_cycles,
            dynamic_amplitude=dynamic_amplitude,
            linear_scale=linear_scale,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_sweep(self):
        """
        get_sweep.
        """
        operation = "get_sweep"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def start_sweep(self, single=False, strict=True):
        """
        start_sweep.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type single: `boolean`
        :param single: Enable single sweep (otherwise loop).

        """
        operation = "start_sweep"
        params = dict(
            strict=strict,
            single=single,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def stop_sweep(self):
        """
        stop_sweep.
        """
        operation = "stop_sweep"
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation)

    def set_output(
        self,
        channel,
        amplitude,
        offset=0,
        enable_amplitude=True,
        enable_offset=True,
        strict=True,
    ):
        """
        set_output.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type amplitude: `number` [-5V, 5V]
        :param amplitude: Waveform peak-to-peak amplitude

        :type offset: `number` [-5V, 5V]  (defaults to 0)
        :param offset: DC offset applied to the waveform

        :type enable_amplitude: `bool`(defaults to True)
        :param enable_amplitude: Enable amplitude on the given channel

        :type enable_offset: `bool`(defaults to True)
        :param enable_offset: Enable offset on the given channel

        """
        operation = "set_output"
        params = dict(
            strict=strict,
            channel=channel,
            amplitude=amplitude,
            offset=offset,
            enable_amplitude=enable_amplitude,
            enable_offset=enable_offset,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output(self, channel):
        """
        get_output.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def disable_output(self, channel, strict=True):
        """
        disable_output.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "disable_output"
        params = dict(
            strict=strict,
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_output_phase(self, channel, phase, strict=True):
        """
        set_output_phase.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type phase: `number`
        :param phase: Output phase

        """
        operation = "set_output_phase"
        params = dict(
            strict=strict,
            channel=channel,
            phase=phase,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_harmonic_multiplier(self, multiplier=1, strict=True):
        """
        set_harmonic_multiplier.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type multiplier: `integer` [1, 15]  (defaults to 1)
        :param multiplier: Multiplier applied to the fundamental frequency

        """
        operation = "set_harmonic_multiplier"
        params = dict(
            strict=strict,
            multiplier=multiplier,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_harmonic_multiplier(self):
        """
        get_harmonic_multiplier.
        """
        operation = "get_harmonic_multiplier"
        return self.session.get(f"slot{self.slot}/{self.operation_group}", operation)

    def set_output_load(self, channel, load, strict=True):
        """
        .. deprecated:: 3.1.1
        Use `set_output_termination` instead.

        set_output_load.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type load: `string` ['1MOhm', '50Ohm']
        :param load: Output load

        """
        operation = "set_output_load"
        params = dict(
            strict=strict,
            channel=channel,
            load=load,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output_load(self, channel):
        """
        .. deprecated:: 3.1.1
        Use `get_output_termination` instead.

        get_output_load.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output_load"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def set_output_termination(self, channel, termination, strict=True):
        """
        set_output_termination.

        :type strict: `boolean`
        :param strict: Disable all implicit conversions and coercions.

        :type channel: `integer`
        :param channel: Target channel

        :type termination: `string` ['HiZ', '50Ohm']
        :param termination: Output termination

        """
        operation = "set_output_termination"
        params = dict(
            strict=strict,
            channel=channel,
            termination=termination,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_output_termination(self, channel):
        """
        get_output_termination.

        :type channel: `integer`
        :param channel: Target channel

        """
        operation = "get_output_termination"
        params = dict(
            channel=channel,
        )
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)

    def get_data(self, timeout=60, wait_reacquire=False, wait_complete=False):
        """
        get_data.

        :type timeout: `number` Seconds (defaults to 60)
        :param timeout: Wait for n seconds to receive a data frame

        :type wait_reacquire: `boolean`
        :param wait_reacquire: Wait until the data is reacquired

        :type wait_complete: `boolean`
        :param wait_complete: Wait until entire frame is available


        .. important::
            Default timeout for reading the data is 10 seconds. It
            can be increased by setting the read_timeout property of
            session object.

            Example: ``i.session.read_timeout=100`` (in seconds)

        """
        operation = "get_data"
        params = dict(timeout=timeout, wait_reacquire=wait_reacquire, wait_complete=wait_complete)
        return self.session.post(f"slot{self.slot}/{self.operation_group}", operation, params)



"""
Logging configuration for the Moku library.

This module sets up logging for the Moku library following best practices
for Python libraries:
- Uses NullHandler by default to avoid unwanted log output
- Allows library users to configure their own logging
- Provides convenience functions for common logging scenarios
"""

import logging
import sys
from typing import Optional

# Create a logger for the moku package
logger = logging.getLogger("moku")

# Add NullHandler to prevent any output by default
# This follows best practice for libraries - let the application control logging
logger.addHandler(logging.NullHandler())

# Set propagation to True so parent loggers can handle our logs
logger.propagate = True


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger for a specific module within the moku package.

    Args:
        name: Module name (e.g., 'session', 'instruments.oscilloscope')

    Returns:
        Logger instance for the specified module
    """
    return logging.getLogger(f"moku.{name}")


def enable_debug_logging(
    level: int = logging.DEBUG, format_string: Optional[str] = None, stream=None
) -> None:
    """
    Enable debug logging for the Moku library.

    This is a convenience function for library users to quickly enable
    debug output. Should only be called by the application, not the library.

    Args:
        level: Logging level (default: DEBUG)
        format_string: Custom format string (default: includes timestamp, level, module, message)
        stream: Output stream (default: stderr)

    Example:
        >>> import moku.logging
        >>> moku.logging.enable_debug_logging()
        >>> # Now all moku operations will log debug information
    """
    if format_string is None:
        format_string = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    if stream is None:
        stream = sys.stderr

    # Remove any existing handlers to avoid duplicates
    for handler in logger.handlers[:]:
        if isinstance(handler, logging.StreamHandler):
            logger.removeHandler(handler)

    # Create and configure handler
    handler = logging.StreamHandler(stream)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)

    # Add handler and set level
    logger.addHandler(handler)
    logger.setLevel(level)


def disable_debug_logging() -> None:
    """
    Disable debug logging for the Moku library.

    Removes stream handlers and resets to NullHandler only.
    """
    # Remove all stream handlers
    for handler in logger.handlers[:]:
        if isinstance(handler, logging.StreamHandler):
            logger.removeHandler(handler)

    # Ensure we have at least a NullHandler
    if not logger.handlers:
        logger.addHandler(logging.NullHandler())

    # Reset level to default
    logger.setLevel(logging.WARNING)


class LoggingContext:
    """
    Context manager for temporarily enabling debug logging.

    Example:
        >>> with moku.logging.LoggingContext():
        ...     # Debug logging enabled here
        ...     moku_device.get_data()
        >>> # Debug logging disabled after context
    """

    def __init__(
        self, level: int = logging.DEBUG, format_string: Optional[str] = None, stream=None
    ):
        self.level = level
        self.format_string = format_string
        self.stream = stream
        self.previous_level = None
        self.previous_handlers = None

    def __enter__(self):
        self.previous_level = logger.level
        self.previous_handlers = logger.handlers[:]
        enable_debug_logging(self.level, self.format_string, self.stream)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore previous state
        logger.handlers = self.previous_handlers
        logger.setLevel(self.previous_level)
        return False



import json
from datetime import datetime
import socket
import time
from subprocess import DEVNULL, PIPE, Popen
from threading import Event, Thread

from moku import MOKU_CLI_PATH
from moku.utilities import check_mokucli_version
from moku.exceptions import StreamException


class MokuCLIThread(Thread):
    """
    Threadinng class responsible to run the mokucli command.

    :param command: Command to execute as a subprocess
    :param start_evt: Event object to set once the subprocess is launched
    """

    def __init__(self, command, error_event, start_evt=None):
        self.command = command
        self.start_evt = start_evt
        self.error_event = error_event
        super(MokuCLIThread, self).__init__()

    def run(self):
        process = Popen(self.command, stdout=DEVNULL, stderr=PIPE)
        # signal subprocess launch...
        if self.start_evt:
            self.start_evt.set()
        process.wait()  # waits until the subprocess is finished executing
        error = process.stderr.read().decode("utf8")
        if error:
            self.error_event.set()
            raise StreamException(error)


class StreamInstrument:
    """
    Base class for all streaming features. Any instrument
    supporting Data streaming should inherit this class
    """

    def __init__(self, mokuOS_version):
        self.stream_id = None
        self.ip_address = None
        self.port = None
        self._socket_rdr = None
        self._running = False
        self._error_event = Event()

    @staticmethod
    def _get_next_available_port():
        sock = socket.socket()
        sock.bind(("", 0))
        port = sock.getsockname()[1]
        sock.close()
        return port

    def _reset_stream_config(self):
        self.stream_id = None
        self.ip_address = None
        self.port = None
        self._socket_rdr = None
        self._running = False
        self._error_event = Event()

    def _connect(self):
        """
        Tries to connect to the tcp port or raises Exception
        """
        i = 0
        while True:
            try:
                i += 1
                if self._error_event.is_set():
                    raise Exception
                client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                client.connect(("localhost", self.port))
                self._socket_rdr = client.makefile("r")
                break
            except socket.error:
                if i == 5:
                    raise Exception(f"Cannot connect to port {self.port}")
                time.sleep(0.5)

    def _begin_streaming(self):
        self.port = self._get_next_available_port()
        _ip = f"--ip-address={self.ip_address}"
        _stream = f"--stream-id={self.stream_id}"
        _target = f"--target={self.port}"
        command = [MOKU_CLI_PATH, "stream", _ip, _stream, _target]
        _start_event = Event()

        cli_thread = MokuCLIThread(command, self._error_event, _start_event)
        cli_thread.start()
        _start_event.wait()
        self._running = True
        self._connect()

    def start_streaming(self):
        """
        Base class start_streaming, verifies if streaming is possible
        """
        check_mokucli_version(MOKU_CLI_PATH)

    def stream_to_file(self, name=None):
        """
        Streams the data to the file of desired format

        :type name: `string`
        :param name: Base name with one of csv, npy, mat extensions (defaults to csv) # noqa

        """
        if not self.stream_id:
            raise StreamException(
                "No streaming session in progress, start one using start_streaming"
            )  # noqa
        if not self._running:
            check_mokucli_version(MOKU_CLI_PATH)
            _ip = f"--ip-address={self.ip_address}"
            _stream = f"--stream-id={self.stream_id}"

            if not name:
                _ts = datetime.now()
                name = f"STREAM_{_ts.strftime('%d%m%Y%H%M%S')}.csv"

            _target = f"--target={name}"
            command = [MOKU_CLI_PATH, "stream", _ip, _stream, _target]
            cli_thread = MokuCLIThread(command, self._error_event)
            cli_thread.start()

    def get_stream_data(self):
        """
        Get the converted stream of data

        :raises StreamException: Indicates END OF STREAM
        """
        if not self.stream_id:
            raise StreamException(
                "No streaming session in progress, start one using start_streaming"
            )  # noqa
        if self._error_event.is_set():
            raise Exception
        if not self._running:
            check_mokucli_version(MOKU_CLI_PATH)
            self._begin_streaming()
        data = self._socket_rdr.readline()
        if data:
            if data == "EOS\n":
                self._reset_stream_config()
                raise StreamException("End of stream")
            return json.loads(data)
